# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
# ================================================================================================
"""
A random sampler for unit testing and debugging.
"""
from random import choice

from dimod.core.sampler import Sampler
from dimod.response import Response, SampleView

__all__ = ['RandomSampler']


class RandomSampler(Sampler):
    """A sampler that gives random samples for testing.

    Examples:
        This example provides random samples for a two-variable QUBO model.

        >>> import dimod
        >>> response = dimod.RandomSampler().sample_qubo({(0, 0): -1, (1, 1): -1, (0, 1): 2}, num_reads=5)
        >>> len(response)
        5
        >>> print(next(response.data()))      # doctest: +SKIP
        Sample(sample={0: 1, 1: 0}, energy=-1.0)

    """
    properties = None
    parameters = None

    def __init__(self):
        self.parameters = {'num_reads': []}
        self.properties = {}

    def sample(self, bqm, num_reads=10):
        """Give random samples for a binary quadratic model.

        Args:
            bqm (:obj:`~dimod.BinaryQuadraticModel`):
                Binary quadratic model to be sampled from.
            num_reads (int, optional):
                Number of reads.

        Returns:
            :obj:`~dimod.Response`: A `dimod` :obj:`.~dimod.Response` object.

        Notes:
            For each variable in each sample, the value is chosen by a coin flip.

        Examples:
            This example provides samples for a two-variable Ising model.

            >>> import dimod
            >>> sampler = dimod.RandomSampler()
            >>> h = {0: -1, 1: -1}
            >>> J = {(0, 1): -1}
            >>> bqm = dimod.BinaryQuadraticModel(h, J, -0.5, dimod.SPIN)
            >>> response = sampler.sample(bqm, num_reads=3)
            >>> len(response)
            3
            >>> response.data_vectors['energy']        # doctest: +SKIP
            array([ 0.5, -3.5,  0.5])

        """
        values = tuple(bqm.vartype.value)

        def _itersample():
            for __ in range(num_reads):
                sample = {v: choice(values) for v in bqm.linear}
                energy = bqm.energy(sample)

                yield sample, energy

        samples, energies = zip(*_itersample())

        return Response.from_dicts(samples, {'energy': energies}, vartype=bqm.vartype)



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
# ================================================================================================
"""
An exact solver that calculates the energy of all possible samples.
"""
import itertools

import numpy as np
from six.moves import zip

from dimod.core.sampler import Sampler
from dimod.decorators import bqm_index_labels
from dimod.response import Response
from dimod.vartypes import Vartype

__all__ = ['ExactSolver']


class ExactSolver(Sampler):
    """A simple exact solver for testing and debugging.

    Notes:
        This solver becomes slow for problems with 18 or more
        variables.

    Examples:
        This example solves a two-variable Ising model.

        >>> import dimod
        >>> response = dimod.ExactSolver().sample_ising({'a': -0.5, 'b': 1.0}, {('a', 'b'): -1})
        >>> response.data_vectors['energy']
        array([-1.5, -0.5, -0.5,  2.5])

    """
    properties = None
    parameters = None

    def __init__(self):
        self.properties = {}
        self.parameters = {}

    @bqm_index_labels
    def sample(self, bqm):
        """Sample from a binary quadratic model.

        Args:
            bqm (:obj:`~dimod.BinaryQuadraticModel`):
                Binary quadratic model to be sampled from.

        Returns:
            :obj:`~dimod.Response`: A `dimod` :obj:`.~dimod.Response` object.


        Examples:
            This example provides samples for a two-variable Ising model.

            >>> import dimod
            >>> sampler = dimod.ExactSolver()
            >>> bqm = dimod.BinaryQuadraticModel({0: 0.0, 1: 1.0}, {(0, 1): 0.5}, -0.5, dimod.SPIN)
            >>> response = sampler.sample(bqm)
            >>> response.data_vectors['energy']
            array([-1., -2.,  1.,  0.])

        """
        M = bqm.binary.to_numpy_matrix()
        off = bqm.binary.offset

        if M.shape == (0, 0):
            return Response.empty(bqm.vartype)

        sample = np.zeros((len(bqm),), dtype=bool)

        # now we iterate, flipping one bit at a time until we have
        # traversed all samples. This is a Gray code.
        # https://en.wikipedia.org/wiki/Gray_code
        def iter_samples():
            sample = np.zeros((len(bqm)), dtype=bool)
            energy = 0.0

            yield sample.copy(), energy + off

            for i in range(1, 1 << len(bqm)):
                v = _ffs(i)

                # flip the bit in the sample
                sample[v] = not sample[v]

                # for now just calculate the energy, but there is a more clever way by calculating
                # the energy delta for the single bit flip, don't have time, pull requests
                # appreciated!
                energy = sample.dot(M).dot(sample.transpose())

                yield sample.copy(), float(energy) + off

        samples, energies = zip(*iter_samples())

        response = Response.from_matrix(np.matrix(samples, dtype='int8'), {'energy': energies},
                                        vartype=Vartype.BINARY)

        # make sure the response matches the given vartype, in-place.
        response.change_vartype(bqm.vartype, inplace=True)

        return response


def _ffs(x):
    """Gets the index of the least significant set bit of x."""
    return (x & -x).bit_length() - 1



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
# ================================================================================================

import re

from dimod.binary_quadratic_model import BinaryQuadraticModel

_LINE_REGEX = r'^\s*(\d+)\s+(\d+)\s+([+-]?([0-9]*[.])?[0-9]+)\s*$'
"""
Each line should look like

0 1 2.000000

where 0, 1 are the variable lables and 2.0 is the bias.
"""


def dumps(bqm):
    """Dump a binary quadratic model to a string in COOrdinate format."""
    return '\n'.join(_iter_triplets(bqm))


def dump(bqm, fp):
    """Dump a binary quadratic model to a string in COOrdinate format."""
    for triplet in _iter_triplets(bqm):
        fp.write('%s\n' % triplet)


def loads(s, bqm):
    """Load a COOrdinate formatted binary quadratic model from a string."""
    pattern = re.compile(_LINE_REGEX)

    for line in s.split('\n'):
        match = pattern.search(line)

        if match is not None:
            u, v, bias = int(match.group(1)), int(match.group(2)), float(match.group(3))
            if u == v:
                bqm.add_variable(u, bias)
            else:
                bqm.add_interaction(u, v, bias)

    return bqm


def load(fp, bqm):
    """Load a COOrdinate formatted binary quadratic model from a file."""
    pattern = re.compile(_LINE_REGEX)

    for line in fp:
        match = pattern.search(line)

        if match is not None:
            u, v, bias = int(match.group(1)), int(match.group(2)), float(match.group(3))
            if u == v:
                bqm.add_variable(u, bias)
            else:
                bqm.add_interaction(u, v, bias)

    return bqm


def _iter_triplets(bqm):
    if not isinstance(bqm, BinaryQuadraticModel):
        raise TypeError("expected input to be a BinaryQuadraticModel")
    if not all(isinstance(v, int) and v >= 0 for v in bqm.linear):
        raise ValueError("only positive index-labeled binary quadratic models can be dumped to COOrdinate format")

    # developer note: we could (for some threshold sparseness) sort the neighborhoods,
    # but this is simple and probably sufficient

    variables = sorted(bqm)  # integer labeled so we can sort in py3

    for idx, u in enumerate(variables):
        for v in variables[idx:]:
            if u == v and bqm.linear[u]:
                yield '%d %d %f' % (u, u, bqm.linear[u])
            elif u in bqm.adj[v]:
                yield '%d %d %f' % (u, v, bqm.adj[u][v])



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
# ================================================================================================


class InvalidComposition(Exception):
    """Raised for compositions of samplers that are invalid"""


class MappingError(Exception):
    """Raised when mapping causes conflicting values in samples"""


class InvalidSampler(Exception):
    """Raised when trying to use the specified sampler as a sampler"""


class BinaryQuadraticModelValueError(ValueError):
    """Raised when a sampler cannot handle a specified binary quadratic model"""


class BinaryQuadraticModelSizeError(BinaryQuadraticModelValueError):
    """Raised when the binary quadratic model has too many variables"""


class BinaryQuadraticModelStructureError(BinaryQuadraticModelValueError):
    """Raised when the binary quadratic model does not fit the sampler"""



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
# ================================================================================================

import inspect
from functools import wraps

from six import iteritems

from dimod.compatibility23 import getargspec
from dimod.core.structured import Structured
from dimod.exceptions import BinaryQuadraticModelStructureError
from dimod.vartypes import Vartype

__all__ = ['bqm_index_labels', 'bqm_structured', 'mixinmethod', 'vartype_argument']


def bqm_index_labels(f):
    """todo
    """

    @wraps(f)
    def _index_label(sampler, bqm, **kwargs):
        if not hasattr(bqm, 'linear'):
            raise TypeError('expected input to be a BinaryQuadraticModel')
        linear = bqm.linear

        # if already index-labelled, just continue
        if all(v in linear for v in range(len(bqm))):
            return f(sampler, bqm, **kwargs)

        try:
            inverse_mapping = dict(enumerate(sorted(linear)))
        except TypeError:
            # in python3 unlike types cannot be sorted
            inverse_mapping = dict(enumerate(linear))
        mapping = {v: i for i, v in iteritems(inverse_mapping)}

        response = f(sampler, bqm.relabel_variables(mapping, inplace=False), **kwargs)

        # unapply the relabeling
        return response.relabel_variables(inverse_mapping, inplace=True)

    return _index_label


def bqm_structured(f):
    """todo

    makes sure bqm has the appropriate structure
    """

    @wraps(f)
    def new_f(sampler, bqm, **kwargs):
        try:
            structure = sampler.structure
            adjacency = structure.adjacency
        except AttributeError:
            if isinstance(sampler, Structured):
                raise RuntimeError("something is wrong with the structured sampler")
            else:
                raise TypeError("sampler does not have a structure property")

        if not all(v in adjacency for v in bqm.linear):
            # todo: better error message
            raise BinaryQuadraticModelStructureError("given bqm does not match the sampler's structure")
        if not all(u in adjacency[v] for u, v in bqm.quadratic):
            # todo: better error message
            raise BinaryQuadraticModelStructureError("given bqm does not match the sampler's structure")

        return f(sampler, bqm, **kwargs)

    return new_f


def vartype_argument(*arg_names):
    """Ensures the wrapped function receives valid vartype argument(s). One
    or more argument names can be specified (as a list of string arguments).

    Args:
        *arg_names (list[str], argument names, optional, default='vartype'):
            The names of the constrained arguments in function decorated.

    Returns:
        Function decorator.

    Examples:
        >>> @dimod.vartype_argument()
        ... def f(x, vartype):
        ...     print(vartype)
        ...
        >>> f(1, 'SPIN')
        Vartype.SPIN
        >>> f(1, vartype='SPIN')
        Vartype.SPIN

        >>> @dimod.vartype_argument('y')
        ... def f(x, y):
        ...     print(y)
        ...
        >>> f(1, 'SPIN')
        Vartype.SPIN
        >>> f(1, y='SPIN')
        Vartype.SPIN

        >>> @dimod.vartype_argument('z')
        ... def f(x, **kwargs):
        ...     print(kwargs['z'])
        ...
        >>> f(1, z='SPIN')
        Vartype.SPIN

    Note:
        The function decorated can explicitly list (name) vartype arguments
        constrained by :func:`vartype_argument`, or it can use a keyword arguments `dict`.
    """
    # by default, constrain only one argument, the 'vartype`
    if not arg_names:
        arg_names = ['vartype']

    def _vartype_arg(f):
        argspec = getargspec(f)

        def _enforce_single_arg(name, args, kwargs):
            try:
                vartype = kwargs[name]
            except KeyError:
                raise TypeError('vartype argument missing')

            if isinstance(vartype, Vartype):
                return

            try:
                if isinstance(vartype, str):
                    vartype = Vartype[vartype]
                else:
                    vartype = Vartype(vartype)

            except (ValueError, KeyError):
                raise TypeError(("expected input vartype to be one of: "
                                 "Vartype.SPIN, 'SPIN', {-1, 1}, "
                                 "Vartype.BINARY, 'BINARY', or {0, 1}."))

            kwargs[name] = vartype

        @wraps(f)
        def new_f(*args, **kwargs):
            # bound actual f arguments (including defaults) to f argument names
            # (note: if call arguments don't match actual function signature,
            # we'll fail here with the standard `TypeError`)
            bound_args = inspect.getcallargs(f, *args, **kwargs)

            # `getcallargs` doesn't merge additional positional/keyword arguments,
            # so do it manually
            final_args = list(bound_args.pop(argspec.varargs, ()))
            final_kwargs = bound_args.pop(argspec.keywords, {})

            final_kwargs.update(bound_args)
            for name in arg_names:
                _enforce_single_arg(name, final_args, final_kwargs)

            return f(*final_args, **final_kwargs)

        return new_f

    return _vartype_arg



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
# ================================================================================================

import itertools

from collections import Counter

from six import iteritems

from dimod.binary_quadratic_model import BinaryQuadraticModel
from dimod.decorators import vartype_argument
from dimod.vartypes import Vartype

__all__ = ['make_quadratic']


def _spin_product(variables):
    """Create a bqm with a gap of 2 that represents the product of two variables.

    Note that spin-product requires an auxiliary variable.

    Args:
        variables (list):
            multiplier, multiplicand, product, aux

    Returns:
        :obj:`.BinaryQuadraticModel`

    """
    multiplier, multiplicand, product, aux = variables

    return BinaryQuadraticModel({multiplier: -.5,
                                 multiplicand: -.5,
                                 product: -.5,
                                 aux: -1.},
                                {(multiplier, multiplicand): .5,
                                 (multiplier, product): .5,
                                 (multiplier, aux): 1.,
                                 (multiplicand, product): .5,
                                 (multiplicand, aux): 1.,
                                 (product, aux): 1.},
                                2.,
                                Vartype.SPIN)


def _binary_product(variables):
    """Create a bqm with a gap of 2 that represents the product of two variables.

    Args:
        variables (list):
            multiplier, multiplicand, product

    Returns:
        :obj:`.BinaryQuadraticModel`

    """
    multiplier, multiplicand, product = variables

    return BinaryQuadraticModel({multiplier: 0.0,
                                 multiplicand: 0.0,
                                 product: 3.0},
                                {(multiplier, multiplicand): 1.0,
                                 (multiplier, product): -2.0,
                                 (multiplicand, product): -2.0},
                                0.0,
                                Vartype.BINARY)


def make_quadratic(poly, strength, vartype=None, bqm=None):
    """Create a binary quadratic model from a higher order polynomial.

    Args:
        poly (dict):
            A polynomial. Should be a dict of the form {term: bias, ...} where term is a tuple of
            variables and bias is their associated bias.

        strength (float):
            The strength of the reduction constraint. Insufficient strength can result in the
            binary quadratic model not having the same minimizations as the polynomial.

        vartype (:class:`.Vartype`, optional):
            The vartype of the polynomial. If a bqm is provided, then vartype is not required.

        bqm (:class:`.BinaryQuadraticModel`, optional):
            The terms of the reduced polynomial are added to this bqm. If not provided a new
            empty binary quadratic model is created.

    Returns:
        :class:`.BinaryQuadraticModel`

    Examples:

        >>> import dimod
        ...
        >>> poly = {(0,): -1, (1,): 1, (2,): 1.5, (0, 1): -1, (0, 1, 2): -2}
        >>> bqm = make_quadratic(poly, 5.0, dimod.SPIN)

    """

    if bqm is None:
        if vartype is None:
            raise ValueError("one of vartype and create_using must be provided")
        bqm = BinaryQuadraticModel.empty(vartype)
    else:
        if not isinstance(bqm, BinaryQuadraticModel):
            raise TypeError('create_using must be a BinaryQuadraticModel')
        if vartype is not None and vartype is not bqm.vartype:
            raise ValueError("one of vartype and create_using must be provided")
    bqm.info['reduction'] = {}

    new_poly = {}
    for term, bias in iteritems(poly):
        if len(term) == 0:
            bqm.add_offset(bias)
        elif len(term) == 1:
            v, = term
            bqm.add_variable(v, bias)
        else:
            new_poly[term] = bias

    return _reduce_degree(bqm, new_poly, vartype, strength)


def _reduce_degree(bqm, poly, vartype, scale):

    if all(len(term) <= 2 for term in poly):
        # termination criteria, we are already quadratic
        bqm.add_interactions_from(poly)
        return bqm

    # determine which pair of variables appear most often
    paircounter = Counter()
    for term in poly:
        if len(term) > 2:
            for u, v in itertools.combinations(term, 2):
                pair = frozenset((u, v))
                paircounter[pair] += 1

    pair, __ = paircounter.most_common(1)[0]
    u, v = pair

    # make a new product variable and aux variable and add constraint that u*v == p
    p = '{}*{}'.format(u, v)

    while p in bqm.linear:
        p = '_' + p

    if vartype is Vartype.BINARY:
        constraint = _binary_product([u, v, p])

        bqm.info['reduction'][(u, v)] = {'product': p}
    else:
        aux = 'aux{},{}'.format(u, v)
        while aux in bqm.linear:
            aux = '_' + aux
        constraint = _spin_product([u, v, p, aux])

        bqm.info['reduction'][(u, v)] = {'product': p, 'auxiliary': aux}

    constraint.scale(scale)
    bqm.update(constraint)

    new_poly = {}
    for interaction, bias in poly.items():
        if u in interaction and v in interaction:

            if len(interaction) == 2:
                # in this case we are reducing a quadratic bias, so it becomes linear and can
                # be removed
                assert len(interaction) >= 2
                bqm.add_variable(p, bias)
                continue

            interaction = tuple(s for s in interaction if s not in pair)
            interaction += (p,)

        if interaction in new_poly:
            new_poly[interaction] += bias
        else:
            new_poly[interaction] = bias

    return _reduce_degree(bqm, new_poly, vartype, scale)



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
# ================================================================================================
"""
A composite that structures a sampler.
"""
from dimod.core.sampler import Sampler
from dimod.core.composite import Composite
from dimod.core.structured import Structured
from dimod.decorators import bqm_structured


class StructureComposite(Sampler, Composite, Structured):
    """Creates a structured composed sampler from an unstructured sampler.

    Args:
        Sampler (:obj:`~dimod.Sampler`):
            Unstructured sampler.
        nodelist (list):
            Nodes/variables allowed by the sampler formatted as a list.
        edgelist (list[(node, node)]):
            Edges/interactions allowed by the sampler, formatted as a list where each
            edge/interaction is a 2-tuple.

    Examples:
        This example creates a composed sampler from the unstructure dimod ExactSolver sampler.
        The target structure is a square graph.

        >>> import dimod
        >>> base_sampler = dimod.ExactSolver()
        >>> node_list = [0, 1, 2, 3]
        >>> edge_list = [(0, 1), (1, 2), (2, 3), (0, 3)]
        >>> structured_sampler = dimod.StructureComposite(base_sampler, node_list, edge_list)
        >>> linear = {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0}
        >>> quadratic = {(0, 1): 1.0, (1, 2): 1.0, (0, 3): 1.0, (2, 3): -1.0}
        >>> bqm = dimod.BinaryQuadraticModel(linear, quadratic, 1.0, dimod.Vartype.SPIN)
        >>> response = structured_sampler.sample(bqm)
        >>> print(next(response.data()))
        Sample(sample={0: 1, 1: -1, 2: -1, 3: -1}, energy=-1.0)
        >>> # Try giving the composed sampler a non-square model
        >>> del quadratic[(0, 1)]
        >>> quadratic[(0, 2)] = 1.0
        >>> bqm = dimod.BinaryQuadraticModel(linear, quadratic, 1.0, dimod.Vartype.SPIN)
        >>> try: response = structured_sampler.sample(bqm)    # doctest: +SKIP
        ... except dimod.BinaryQuadraticModelStructureError as details:
        ...     print(details)
        ...
        given bqm does not match the sampler's structure

    """
    # we will override these in the __init__, but because they are abstract properties we need to
    # signal that we are overriding them
    edgelist = None
    nodelist = None
    children = None

    def __init__(self, sampler, nodelist, edgelist):
        self.children = [sampler]
        self.nodelist = nodelist
        self.edgelist = edgelist

    @property
    def parameters(self):
        return self.child.parameters

    @property
    def properties(self):
        return self.child.properties

    @bqm_structured
    def sample(self, bqm, **sample_kwargs):
        """Sample from the binary quadratic model.

        Args:
            bqm (:obj:`~dimod.BinaryQuadraticModel`):
                Binary quadratic model to be sampled from.

        Returns:
            :obj:`~dimod.Response`: A `dimod` :obj:`.~dimod.Response` object.

        Examples:
            This example submits an Ising problem to a composed sampler that uses
            the dimod ExactSampler only on problems structured for a K2 fully connected graph.

            >>> import dimod
            >>> response = dimod.StructureComposite(dimod.ExactSolver(),
            ...                  [0, 1], [(0, 1)]).sample_ising({0: 1, 1: 1}, {})
            >>> print(next(response.data()))
            Sample(sample={0: -1, 1: -1}, energy=-2.0)
        """
        return self.child.sample(bqm, **sample_kwargs)



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
# ================================================================================================

from __future__ import division, absolute_import

from six import iteritems
import numpy as np

from dimod.response import Response


__all__ = ['target_to_source', 'chain_to_quadratic', 'chain_break_frequency']


def target_to_source(target_adjacency, embedding):
    """Derive the source adjacency from an embedding and target adjacency.

    Args:
        target_adjacency (dict/:class:`networkx.Graph`):
            A dict of the form {v: Nv, ...} where v is a node in the target graph and Nv is the
            neighbors of v as an iterable. This can also be a networkx graph.

        embedding (dict):
            A mapping from a source graph to a target graph.

    Returns:
        dict: The adjacency of the source graph.

    Raises:
        ValueError: If any node in the target_adjacency is assigned more
            than  one node in the source graph by embedding.

    Examples:

        >>> target_adjacency = {0: {1, 3}, 1: {0, 2}, 2: {1, 3}, 3: {0, 2}}  # a square graph
        >>> embedding = {'a': {0}, 'b': {1}, 'c': {2, 3}}
        >>> source_adjacency = dimod.embedding.target_to_source(target_adjacency, embedding)
        >>> source_adjacency  # triangle
        {'a': {'b', 'c'}, 'b': {'a', 'c'}, 'c': {'a', 'b'}}

        This function also works with networkx graphs.

        >>> import networkx as nx
        >>> target_graph = nx.complete_graph(5)
        >>> embedding = {'a': {0, 1, 2}, 'b': {3, 4}}
        >>> dimod.embedding.target_to_source(target_graph, embedding)

    """
    # the nodes in the source adjacency are just the keys of the embedding
    source_adjacency = {v: set() for v in embedding}

    # we need the mapping from each node in the target to its source node
    reverse_embedding = {}
    for v, chain in iteritems(embedding):
        for u in chain:
            if u in reverse_embedding:
                raise ValueError("target node {} assigned to more than one source node".format(u))
            reverse_embedding[u] = v

    # v is node in target, n node in source
    for v, n in iteritems(reverse_embedding):
        neighbors = target_adjacency[v]

        # u is node in target
        for u in neighbors:

            # some nodes might not be assigned to chains
            if u not in reverse_embedding:
                continue

            # m is node in source
            m = reverse_embedding[u]

            if m == n:
                continue

            source_adjacency[n].add(m)
            source_adjacency[m].add(n)

    return source_adjacency


def chain_to_quadratic(chain, target_adjacency, chain_strength):
    """Determine the quadratic biases that induce the given chain.

    Args:
        chain (iterable):
            The variables that make up a chain.

        target_adjacency (dict/:class:`networkx.Graph`):
            Should be a dict of the form {s: Ns, ...} where s is a variable
            in the target graph and Ns is the set of neighbours of s.

        chain_strength (float):
            The magnitude of the quadratic bias that should be used to create chains.

    Returns:
        dict[edge, float]: The quadratic biases that induce the given chain.

    Raises:
        ValueError: If the variables in chain do not form a connected subgraph of target.

    Examples:
        >>> chain = {1, 2}
        >>> target_adjacency = {0: {1, 2}, 1: {0, 2}, 2: {0, 1}}
        >>> dimod.embedding.chain_to_quadratic(chain, target_adjacency, 1)
        {(1, 2): -1}

    """
    quadratic = {}  # we will be adding the edges that make the chain here

    # do a breadth first search
    seen = set()
    try:
        next_level = {next(iter(chain))}
    except StopIteration:
        raise ValueError("chain must have at least one variable")
    while next_level:
        this_level = next_level
        next_level = set()
        for v in this_level:
            if v not in seen:
                seen.add(v)

                for u in target_adjacency[v]:
                    if u not in chain:
                        continue
                    next_level.add(u)
                    if u != v and (u, v) not in quadratic:
                        quadratic[(v, u)] = -chain_strength

    if len(chain) != len(seen):
        raise ValueError('{} is not a connected chain'.format(chain))

    return quadratic


def chain_break_frequency(samples, embedding):
    """Determine the frequency of chain breaks in the given samples.

    Args:
        samples (array-like/:obj:`.Response`):
            Matrix of samples or a dimod response object.

        embedding (dict):
            Mapping from source graph to target graph as a dict of form {s: {t, ...}, ...},
            where s is a source-model variable and t is a target-model variable.

    Returns:
        dict: Frequency of chain breaks as a dict in the form {s: f, ...},  where s
        is a variable in the source graph, and frequency, a float, is the fraction
        of broken chains.

    Examples:
        This example embeds a single source node, 'a', as a chain of two target nodes (0, 1)
        and uses :func:`.chain_break_frequency` to show that out of two synthetic samples,
        one ([-1, +1]) represents a broken chain.

        >>> import dimod
        >>> import numpy as np
        >>> samples = np.matrix([[-1, +1], [+1, +1]])
        >>> embedding = {'a': {0, 1}}
        >>> print(dimod.chain_break_frequency(samples, embedding)['a'])
        0.5


        This example embeds a single source node (0) as a chain of two target nodes (a, b)
        and uses :func:`.chain_break_frequency` to show that out of two samples in a
        dimod response, one ({'a': 1, 'b': 0}) represents a broken chain.

        >>> import dimod
        >>> response = dimod.Response.from_dicts([{'a': 1, 'b': 0}, {'a': 0, 'b': 0}], {'energy': [1, 0]})
        >>> embedding = {0: {'a', 'b'}}
        >>> print(dimod.chain_break_frequency(response, embedding)[0])
        0.5

    """
    if isinstance(samples, Response):
        if samples.variable_labels is not None:
            label_to_idx = samples.label_to_idx
            embedding = {v: {label_to_idx[u] for u in chain} for v, chain in iteritems(embedding)}
        samples = samples.samples_matrix
    else:
        samples = np.matrix(samples)

    f = {}
    for v, chain in iteritems(embedding):
        chain = list(chain)
        u = chain[0]
        f[v] = 1.0 - float(np.mean((samples[:, chain] == samples[:, u]).all(axis=1), dtype=float))

    return f


def edgelist_to_adjacency(edgelist):
    """Converts an iterator of edges to an adjacency dict.

    Args:
        edgelist (iterable):
            An iterator over 2-tuples where each 2-tuple is an edge.

    Returns:
        dict: The adjacency dict. A dict of the form {v: Nv, ...} where v is a node in a graph and
        Nv is the neighbors of v as an set.

    """
    adjacency = dict()
    for u, v in edgelist:
        if u in adjacency:
            adjacency[u].add(v)
        else:
            adjacency[u] = {v}
        if v in adjacency:
            adjacency[v].add(u)
        else:
            adjacency[v] = {u}
    return adjacency



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
# ================================================================================================
"""Chain-break-resolution generators available to :func:`.iter_unembed`.

Chain-break-resolution generators enable :func:`.iter_unembed` to use different techniques
of resolving chain breaks without keeping large numbers of samples in memory. Each generator
yields zero or more unembedded samples.

"""

from collections import Counter, defaultdict, Callable

from six import iteritems, itervalues

import numpy as np

__all__ = ['majority_vote', 'discard', 'weighted_random', 'MinimizeEnergy']


def discard_matrix(samples_matrix, chain_list):
    """Discard broken chains."""
    if not isinstance(samples_matrix, np.matrix):
        samples_matrix = np.matrix(samples_matrix, dtype='int8')
    num_samples, __ = samples_matrix.shape

    variables = []
    for chain in chain_list:
        chain = list(chain)

        try:
            v = chain[0]
        except IndexError:
            raise ValueError("each chain in chain_list must contain at least one variable")
        variables.append(v)

        if len(chain) == 1:
            continue

        # there must be a better way
        unbroken = np.array((samples_matrix[:, chain] == samples_matrix[:, v]).all(axis=1)).flatten()

        samples_matrix = samples_matrix[unbroken, :]

    return samples_matrix[:, variables]


def discard(sample, embedding):
    """Discard the sample if a chain is broken.

    Args:
        sample (Mapping):
            Sample as a dict of form {t: val, ...}, where t is
            a variable in the target graph and val its associated value as
            determined by a binary quadratic model sampler.
        embedding (dict):
            Mapping from source graph to target graph as a dict
            of form {s: {t, ...}, ...}, where s is a source-model variable and t is
            a target-model variable.

    Yields:
        dict: The unembedded sample if no chains were broken.

    Examples:
        This example unembeds a sample from a target graph that chains nodes 0 and 1 to
        represent source node a. The first sample has an unbroken chain, the second a broken
        chain.

        >>> import dimod
        >>> embedding = {'a': {0, 1}, 'b': {2}}
        >>> samples = {0: 1, 1: 1, 2: 0}
        >>> next(dimod.embedding.discard(samples, embedding), 'No sample')  # doctest: +SKIP
        {'a': 1, 'b': 0}
        >>> samples = {0: 1, 1: 0, 2: 0}
        >>> next(dimod.embedding.discard(samples, embedding), 'No sample')
        'No sample'

    """
    unembedded = {}

    for v, chain in iteritems(embedding):
        vals = [sample[u] for u in chain]

        if _all_equal(vals):
            unembedded[v] = vals.pop()
        else:
            return

    yield unembedded


def majority_vote(sample, embedding):
    """Determine the sample values for chains by majority vote.

    Args:
        sample (Mapping):
            Sample as a dict of form {t: val, ...}, where t is
            a variable in the target graph and val its associated value as
            determined by a binary quadratic model sampler.
        embedding (dict):
            Mapping from source graph to target graph as a dict
            of form {s: {t, ...}, ...}, where s is a source-model variable and t is
            a target-model variable.

    Yields:
        dict: The unembedded sample. When there is a chain break, the value
        is chosen to match the most common value in the chain. For broken chains
        without a majority, one of the two values is chosen arbitrarily.

    Examples:
        This example unembeds a sample from a target graph that chains nodes 0 and 1 to
        represent source node a and nodes 2, 3, and 4 to represent source node b.
        Both samples have broken chains for source node b, with different majority values.

        >>> import dimod
        >>> embedding = {'a': {0, 1}, 'b': {2, 3, 4}}
        >>> samples = {0: 1, 1: 1, 2: 0, 3: 0, 4: 1}
        >>> next(dimod.embedding.majority_vote(samples, embedding), 'No sample')['b']
        0
        >>> samples = {0: 1, 1: 1, 2: 1, 3: 0, 4: 1}
        >>> next(dimod.embedding.majority_vote(samples, embedding), 'No sample')['b']
        1

    """
    unembedded = {}

    for v, chain in iteritems(embedding):
        vals = [sample[u] for u in chain]

        if _all_equal(vals):
            unembedded[v] = vals.pop()
        else:
            unembedded[v] = _most_common(vals)

    yield unembedded


def weighted_random(sample, embedding):
    """Determine the sample values of chains by weighed random choice.

    Args:
        sample (Mapping):
            Sample as a dict of form {t: val, ...}, where t is
            a variable in the target graph and val its associated value as
            determined by a binary quadratic model sampler.
        embedding (dict):
            Mapping from source graph to target graph as a dict
            of form {s: {t, ...}, ...}, where s is a source-model variable and t is
            a target-model variable.

    Yields:
        dict: The unembedded sample. When there is a chain break, the value
        is chosen randomly, weighted by frequency of the chain's values.

    Examples:
        This example unembeds a sample from a target graph that chains nodes 0 and 1 to
        represent source node a and nodes 2, 3, and 4 to represent source node b.
        The sample has broken chains for both source nodes.

        >>> import dimod
        >>> embedding = {'a': {0, 1}, 'b': {2, 3, 4}}
        >>> samples = {0: 1, 1: 0, 2: 1, 3: 0, 4: 1}
        >>> next(dimod.embedding.weighted_random(samples, embedding), 'No sample')  # doctest: +SKIP
        {'a': 0, 'b': 1}

    """
    unembedded = {}

    for v, chain in iteritems(embedding):
        vals = [sample[u] for u in chain]

        # pick a random element uniformly from all vals, this weights them by
        # the proportion of each
        unembedded[v] = np.random.choice(vals)

    yield unembedded


class MinimizeEnergy(Callable):
    """Determine the sample values of broken chains by minimizing local energy.

    Args:
        linear (dict): Linear biases of the source model as a dict of
            form {s: bias, ...}, where s is a source-model variable
            and bias its associated linear bias.
        quadratic (dict): Quadratic biases of the source model as a dict
            of form {(u, v): bias, ...}, where u, v are source-model variables
            and bias the associated quadratic bias.

    Examples:
        This example embeds from a triangular graph to a square graph,
        chaining target-nodes 2 and 3 to represent source-node c, and unembeds
        using the `MinimizeEnergy` method four synthetic samples. The first two
        sample have unbroken chains, the second two have broken chains.

        >>> import dimod
        >>> h = {'a': 0, 'b': 0, 'c': 0}
        >>> J = {('a', 'b'): 1, ('b', 'c'): 1, ('a', 'c'): 1}
        >>> embedding = {'a': {0}, 'b': {1}, 'c': {2, 3}}
        >>> method = dimod.embedding.MinimizeEnergy(h, J)
        >>> samples = [{0: +1, 1: -1, 2: +1, 3: +1},
        ...            {0: -1, 1: -1, 2: -1, 3: -1},
        ...            {0: -1, 1: -1, 2: +1, 3: -1},
        ...            {0: +1, 1: +1, 2: -1, 3: +1}]
        ...
        >>> for source_sample in dimod.iter_unembed(samples, embedding, chain_break_method=method):  # doctest: +SKIP
        ...     print(source_sample)
        ...
        {'a': 1, 'c': 1, 'b': -1}
        {'a': -1, 'c': -1, 'b': -1}
        {'a': -1, 'c': 1, 'b': -1}
        {'a': 1, 'c': -1, 'b': 1}

    """

    def __init__(self, linear=None, quadratic=None):
        if linear is None and quadratic is None:
            raise TypeError("the minimize_energy method requires either `linear` or `quadratic` keyword arguments")
        self._linear = linear if linear is not None else defaultdict(float)
        self._quadratic = quadratic if quadratic is not None else dict()

    def __call__(self, sample, embedding):
        """
        Args:
            sample (dict): Sample as a dict of form {t: val, ...}, where t is
                a target-graph variable and val its associated value as
                determined by a binary quadratic model sampler.
            embedding (dict): Mapping from source graph to target graph as a
                dict of form {s: {t, ...}, ...} where s is a source-graph node
                and t is a target-graph node.

        Yields:
            dict: The unembedded sample. When there is a chain break, the value
            is chosen to minimize the energy relative to its neighbors.
        """
        unembedded = {}
        broken = {}  # keys are the broken source variables, values are the energy contributions

        vartype = set(itervalues(sample))
        if len(vartype) > 2:
            raise ValueError("sample has more than two different values")

        # first establish the values of all of the unbroken chains
        for v, chain in iteritems(embedding):
            vals = [sample[u] for u in chain]

            if _all_equal(vals):
                unembedded[v] = vals.pop()
            else:
                broken[v] = self._linear[v]  # broken tracks the linear energy

        # now, we want to determine the energy for each of the broken variable
        # as much as we can
        for (u, v), bias in iteritems(self._quadratic):
            if u in unembedded and v in broken:
                broken[v] += unembedded[u] * bias
            elif v in unembedded and u in broken:
                broken[u] += unembedded[v] * bias

        # in order of energy contribution, pick spins for the broken variables
        while broken:
            v = max(broken, key=lambda u: abs(broken[u]))  # biggest energy contribution

            # get the value from vartypes that minimizes the energy
            val = min(vartype, key=lambda b: broken[v] * b)

            # set that value and remove it from broken
            unembedded[v] = val
            del broken[v]

            # add v's energy contribution to all of the nodes it is connected to
            for u in broken:
                if (u, v) in self._quadratic:
                    broken[u] += val * self._quadratic[(u, v)]
                if (v, u) in self._quadratic:
                    broken[u] += val * self._quadratic[(v, u)]

        yield unembedded


def _all_equal(iterable):
    """True if all values in `iterable` are equal, else False."""
    iterator = iter(iterable)
    try:
        first = next(iterator)
    except StopIteration:
        # empty iterable is all equal
        return True
    return all(first == rest for rest in iterator)


def _most_common(iterable):
    """Returns the most common element in `iterable`."""
    counts = Counter(iterable)
    if counts:
        (val, __), = counts.most_common(1)
        return val
    else:
        raise ValueError("iterable must contain at least one value")



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
# ================================================================================================

from __future__ import division

import itertools

from six import iteritems, itervalues

from dimod.binary_quadratic_model import BinaryQuadraticModel
from dimod.embedding.chain_breaks import majority_vote
from dimod.embedding.utils import chain_to_quadratic
from dimod.response import Response
from dimod.vartypes import Vartype

__all__ = ['embed_bqm', 'embed_ising', 'embed_qubo', 'iter_unembed', 'unembed_response']


def embed_bqm(source_bqm, embedding, target_adjacency, chain_strength=1.0):
    """Embed a binary quadratic model onto a target graph.

    Args:
        source_bqm (:obj:`.BinaryQuadraticModel`):
            Binary quadratic model to embed.

        embedding (dict):
            Mapping from source graph to target graph as a dict of form {s: {t, ...}, ...},
            where s is a source-model variable and t is a target-model variable.

        target_adjacency (dict/:class:`networkx.Graph`):
            Adjacency of the target graph as a dict of form {t: Nt, ...},
            where t is a variable in the target graph and Nt is its set of neighbours.

        chain_strength (float, optional):
            Magnitude of the quadratic bias (in SPIN-space) applied between variables to create chains. Note
            that the energy penalty of chain breaks is 2 * `chain_strength`.

    Returns:
        :obj:`.BinaryQuadraticModel`: Target binary quadratic model.

    Examples:
        This example embeds a fully connected :math:`K_3` graph onto a square target graph.
        Embedding is accomplished by an edge contraction operation on the target graph:
        target-nodes 2 and 3 are chained to represent source-node c.

        >>> import dimod
        >>> import networkx as nx
        >>> # Binary quadratic model for a triangular source graph
        >>> bqm = dimod.BinaryQuadraticModel.from_ising({}, {('a', 'b'): 1, ('b', 'c'): 1, ('a', 'c'): 1})
        >>> # Target graph is a graph
        >>> target = nx.cycle_graph(4)
        >>> # Embedding from source to target graphs
        >>> embedding = {'a': {0}, 'b': {1}, 'c': {2, 3}}
        >>> # Embed the BQM
        >>> target_bqm = dimod.embed_bqm(bqm, embedding, target)
        >>> target_bqm.quadratic[(0, 1)] == bqm.quadratic[('a', 'b')]
        True
        >>> target_bqm.quadratic   # doctest: +SKIP
        {(0, 1): 1.0, (0, 3): 1.0, (1, 2): 1.0, (2, 3): -1.0}

        This example embeds a fully connected :math:`K_3` graph onto the target graph
        of a dimod reference structured sampler, `StructureComposite`, using the dimod reference
        `ExactSolver` sampler with a square graph specified. Target-nodes 2 and 3
        are chained to represent source-node c.

        >>> import dimod
        >>> # Binary quadratic model for a triangular source graph
        >>> bqm = dimod.BinaryQuadraticModel.from_ising({}, {('a', 'b'): 1, ('b', 'c'): 1, ('a', 'c'): 1})
        >>> # Structured dimod sampler with a structure defined by a square graph
        >>> sampler = dimod.StructureComposite(dimod.ExactSolver(), [0, 1, 2, 3], [(0, 1), (1, 2), (2, 3), (0, 3)])
        >>> # Embedding from source to target graph
        >>> embedding = {'a': {0}, 'b': {1}, 'c': {2, 3}}
        >>> # Embed the BQM
        >>> target_bqm = dimod.embed_bqm(bqm, embedding, sampler.adjacency)
        >>> # Sample
        >>> response = sampler.sample(target_bqm)
        >>> response.samples_matrix   # doctest: +SKIP
        matrix([[-1, -1, -1, -1],
                [ 1, -1, -1, -1],
                [ 1,  1, -1, -1],
                [-1,  1, -1, -1],
                [-1,  1,  1, -1],
        >>> # Snipped above response for brevity

    """
    # create a new empty binary quadratic model with the same class as source_bqm
    target_bqm = source_bqm.empty(source_bqm.vartype)

    # add the offset
    target_bqm.add_offset(source_bqm.offset)

    # start with the linear biases, spreading the source bias equally over the target variables in
    # the chain
    for v, bias in iteritems(source_bqm.linear):

        if v in embedding:
            chain = embedding[v]
        else:
            raise ValueError('no embedding provided for source variable {}'.format(v))

        if any(u not in target_adjacency for u in chain):
            raise ValueError('chain variable {} not in target_adjacency'.format(v))

        b = bias / len(chain)

        target_bqm.add_variables_from({u: b for u in chain})

    # next up the quadratic biases, spread the quadratic biases evenly over the available
    # interactions
    for (u, v), bias in iteritems(source_bqm.quadratic):
        available_interactions = {(s, t) for s in embedding[u] for t in embedding[v] if s in target_adjacency[t]}

        if not available_interactions:
            raise ValueError("no edges in target graph between source variables {}, {}".format(u, v))

        b = bias / len(available_interactions)

        target_bqm.add_interactions_from((u, v, b) for u, v in available_interactions)

    for chain in itervalues(embedding):

        # in the case where the chain has length 1, there are no chain quadratic biases, but we
        # none-the-less want the chain variables to appear in the target_bqm
        if len(chain) == 1:
            v, = chain
            target_bqm.add_variable(v, 0.0)
            continue

        quadratic_chain_biases = chain_to_quadratic(chain, target_adjacency, chain_strength)
        target_bqm.add_interactions_from(quadratic_chain_biases, vartype=Vartype.SPIN)  # these are spin

        # add the energy for satisfied chains to the offset
        energy_diff = -sum(itervalues(quadratic_chain_biases))
        target_bqm.add_offset(energy_diff)

    return target_bqm


def embed_ising(souce_h, source_J, embedding, target_adjacency, chain_strength=1.0):
    """Embed an Ising problem onto a target graph.

    Args:
        source_h (dict[variable, bias]/list[bias]):
            Linear biases of the Ising problem. If a list, the list's indices are used as
            variable labels.

        source_J (dict[(variable, variable), bias]):
            Quadratic biases of the Ising problem.

        embedding (dict):
            Mapping from source graph to target graph as a dict of form {s: {t, ...}, ...},
            where s is a source-model variable and t is a target-model variable.

        target_adjacency (dict/:class:`networkx.Graph`):
            Adjacency of the target graph as a dict of form {t: Nt, ...},
            where t is a target-graph variable and Nt is its set of neighbours.

        chain_strength (float, optional):
            Magnitude of the quadratic bias (in SPIN-space) applied between variables to form a chain. Note
            that the energy penalty of chain breaks is 2 * `chain_strength`.

    Returns:
        tuple: A 2-tuple:

            dict[variable, bias]: Linear biases of the target Ising problem.

            dict[(variable, variable), bias]: Quadratic biases of the target Ising problem.

    Examples:
        This example embeds a fully connected :math:`K_3` graph onto a square target graph.
        Embedding is accomplished by an edge contraction operation on the target graph: target-nodes
        2 and 3 are chained to represent source-node c.

        >>> import dimod
        >>> import networkx as nx
        >>> # Ising problem for a triangular source graph
        >>> h = {}
        >>> J = {('a', 'b'): 1, ('b', 'c'): 1, ('a', 'c'): 1}
        >>> # Target graph is a square graph
        >>> target = nx.cycle_graph(4)
        >>> # Embedding from source to target graph
        >>> embedding = {'a': {0}, 'b': {1}, 'c': {2, 3}}
        >>> # Embed the Ising problem
        >>> target_h, target_J = dimod.embed_ising(h, J, embedding, target)
        >>> target_J[(0, 1)] == J[('a', 'b')]
        True
        >>> target_J        # doctest: +SKIP
        {(0, 1): 1.0, (0, 3): 1.0, (1, 2): 1.0, (2, 3): -1.0}

        This example embeds a fully connected :math:`K_3` graph onto the target graph
        of a dimod reference structured sampler, `StructureComposite`, using the dimod reference
        `ExactSolver` sampler with a square graph specified. Target-nodes 2 and 3 are chained to
        represent source-node c.

        >>> import dimod
        >>> # Ising problem for a triangular source graph
        >>> h = {}
        >>> J = {('a', 'b'): 1, ('b', 'c'): 1, ('a', 'c'): 1}
        >>> # Structured dimod sampler with a structure defined by a square graph
        >>> sampler = dimod.StructureComposite(dimod.ExactSolver(), [0, 1, 2, 3], [(0, 1), (1, 2), (2, 3), (0, 3)])
        >>> # Embedding from source to target graph
        >>> embedding = {'a': {0}, 'b': {1}, 'c': {2, 3}}
        >>> # Embed the Ising problem
        >>> target_h, target_J = dimod.embed_ising(h, J, embedding, sampler.adjacency)
        >>> # Sample
        >>> response = sampler.sample_ising(target_h, target_J)
        >>> for sample in response.samples(n=3, sorted_by='energy'):   # doctest: +SKIP
        ...     print(sample)
        ...
        {0: 1, 1: -1, 2: -1, 3: -1}
        {0: 1, 1: 1, 2: -1, 3: -1}
        {0: -1, 1: 1, 2: -1, 3: -1}

    """
    source_bqm = BinaryQuadraticModel.from_ising(souce_h, source_J)
    target_bqm = embed_bqm(source_bqm, embedding, target_adjacency, chain_strength=chain_strength)
    target_h, target_J, __ = target_bqm.to_ising()
    return target_h, target_J


def embed_qubo(source_Q, embedding, target_adjacency, chain_strength=1.0):
    """Embed a QUBO onto a target graph.

    Args:
        source_Q (dict[(variable, variable), bias]):
            Coefficients of a quadratic unconstrained binary optimization (QUBO) model.

        embedding (dict):
            Mapping from source graph to target graph as a dict of form {s: {t, ...}, ...},
            where s is a source-model variable and t is a target-model variable.

        target_adjacency (dict/:class:`networkx.Graph`):
            Adjacency of the target graph as a dict of form {t: Nt, ...},
            where t is a target-graph variable and Nt is its set of neighbours.

        chain_strength (float, optional):
            Magnitude of the quadratic bias (in SPIN-space) applied between variables to form a chain. Note
            that the energy penalty of chain breaks is 2 * `chain_strength`.

    Returns:
        dict[(variable, variable), bias]: Quadratic biases of the target QUBO.

    Examples:
        This example embeds a square source graph onto fully connected :math:`K_5` graph.
        Embedding is accomplished by an edge deletion operation on the target graph: target-node
        0 is not used.

        >>> import dimod
        >>> import networkx as nx
        >>> # QUBO problem for a square graph
        >>> Q = {(1, 1): -4.0, (1, 2): 4.0, (2, 2): -4.0, (2, 3): 4.0,
        ...      (3, 3): -4.0, (3, 4): 4.0, (4, 1): 4.0, (4, 4): -4.0}
        >>> # Target graph is a fully connected k5 graph
        >>> K_5 = nx.complete_graph(5)
        >>> 0 in K_5
        True
        >>> # Embedding from source to target graph
        >>> embedding = {1: {4}, 2: {3}, 3: {1}, 4: {2}}
        >>> # Embed the QUBO
        >>> target_Q = dimod.embed_qubo(Q, embedding, K_5)
        >>> (0, 0) in target_Q
        False
        >>> target_Q     # doctest: +SKIP
        {(1, 1): -4.0,
         (1, 2): 4.0,
         (2, 2): -4.0,
         (2, 4): 4.0,
         (3, 1): 4.0,
         (3, 3): -4.0,
         (4, 3): 4.0,
         (4, 4): -4.0}

        This example embeds a square graph onto the target graph of a dimod reference structured
        sampler, `StructureComposite`, using the dimod reference `ExactSolver` sampler with a
        fully connected :math:`K_5` graph specified.

        >>> import dimod
        >>> import networkx as nx
        >>> # QUBO problem for a square graph
        >>> Q = {(1, 1): -4.0, (1, 2): 4.0, (2, 2): -4.0, (2, 3): 4.0,
        ...      (3, 3): -4.0, (3, 4): 4.0, (4, 1): 4.0, (4, 4): -4.0}
        >>> # Structured dimod sampler with a structure defined by a K5 graph
        >>> sampler = dimod.StructureComposite(dimod.ExactSolver(), list(K_5.nodes), list(K_5.edges))
        >>> sampler.adjacency      # doctest: +SKIP
        {0: {1, 2, 3, 4},
         1: {0, 2, 3, 4},
         2: {0, 1, 3, 4},
         3: {0, 1, 2, 4},
         4: {0, 1, 2, 3}}
        >>> # Embedding from source to target graph
        >>> embedding = {0: {4}, 1: {3}, 2: {1}, 3: {2}}
        >>> # Embed the QUBO
        >>> target_Q = dimod.embed_qubo(Q, embedding, sampler.adjacency)
        >>> # Sample
        >>> response = sampler.sample_qubo(target_Q)
        >>> for datum in response.data():   # doctest: +SKIP
        ...     print(datum)
        ...
        Sample(sample={1: 0, 2: 1, 3: 1, 4: 0}, energy=-8.0)
        Sample(sample={1: 1, 2: 0, 3: 0, 4: 1}, energy=-8.0)
        Sample(sample={1: 1, 2: 0, 3: 0, 4: 0}, energy=-4.0)
        Sample(sample={1: 1, 2: 1, 3: 0, 4: 0}, energy=-4.0)
        Sample(sample={1: 0, 2: 1, 3: 0, 4: 0}, energy=-4.0)
        Sample(sample={1: 1, 2: 1, 3: 1, 4: 0}, energy=-4.0)
        >>> # Snipped above response for brevity

    """
    source_bqm = BinaryQuadraticModel.from_qubo(source_Q)
    target_bqm = embed_bqm(source_bqm, embedding, target_adjacency, chain_strength=chain_strength)
    target_Q, __ = target_bqm.to_qubo()
    return target_Q


def iter_unembed(target_samples, embedding, chain_break_method=None):
    """Yield unembedded samples.

    :func:`.iter_unembed` is an iterator (to reduce memory footprint) used by :func:`.unembed_response`
    for unembedding; you may use it directly, for example, to increase performance for responses with
    large numbers of samples of which you need only a small portion, say those with lowest energy.

    Args:
        target_samples (iterable[mapping[variable, value]]):
            Iterable of samples. Each sample is a dict of form {t: val, ...},
            where t is a variable in the target graph and val its associated value.

        embedding (dict):
            Mapping from source graph to target graph as a dict of form {s: {t, ...}, ...},
            where s is a source-model variable and t is a target-model variable.

        chain_break_method (function, optional, default=:func:`.majority_vote`):
            Method used to resolve chain breaks.

    Yields:
        dict[variable, value]: An unembedded sample as a dict of form {s: val, ...},
        where s is a variable in the source graph and val its associated value.

    Examples:
        This example demonstrates the use of :func:`.iter_unembed` to derive samples
        for a triangular source graph from synthetic samples of a square target graph.

        >>> import dimod
        >>> # Synthetic samples from a square-structured target problem
        >>> samples = [{0: +1, 1: -1, 2: +1, 3: +1},
        ...            {0: -1, 1: -1, 2: -1, 3: -1},
        ...            {0: +1, 1: +1, 2: +1, 3: +1}]
        >>> # Embedding from source to target graphs
        >>> embedding = {'a': {0}, 'b': {1}, 'c': {2, 3}}
        >>> for source_sample in dimod.iter_unembed(samples, embedding):  # doctest: +SKIP
        ...     print(source_sample)
        ...
        {'a': 1, 'b': -1, 'c': 1}
        {'a': -1, 'b': -1, 'c': -1}
        {'a': 1, 'b': 1, 'c': 1}

        This example uses :func:`.iter_unembed` to unembed samples of a binary quadratic model
        for a triangular source graph that is embedded in a square-structured graph with dimod
        reference structured sampler, `StructureComposite`, using the dimod reference `ExactSolver`
        sampler. Note that this flow is used by dwave-system_'s :obj:`~dwave.system.composites.EmbeddingComposite`.

        >>> import dimod
        >>> # Binary quadratic model for a triangular source graph
        >>> bqm = dimod.BinaryQuadraticModel.from_ising({}, {('a', 'b'): 1, ('b', 'c'): 1, ('a', 'c'): 1})
        >>> # Structured dimod sampler with a structure defined by a square graph
        >>> sampler = dimod.StructureComposite(dimod.ExactSolver(), [0, 1, 2, 3], [(0, 1), (1, 2), (2, 3), (0, 3)])
        >>> # Embedding from source to target graph
        >>> embedding = {'a': {0}, 'b': {1}, 'c': {2, 3}}
        >>> # Embed the BQM
        >>> target_bqm = dimod.embed_bqm(bqm, embedding, sampler.adjacency)
        >>> # Sample (the response is in the form of a target structure)
        >>> target_response = sampler.sample(target_bqm)
        >>> for datum in target_response.data():    # doctest: +SKIP
        ...     print(datum)
        ...
        Sample(sample={0: 1, 1: -1, 2: -1, 3: -1}, energy=-1.0)
        Sample(sample={0: 1, 1: 1, 2: -1, 3: -1}, energy=-1.0)
        Sample(sample={0: -1, 1: 1, 2: -1, 3: -1}, energy=-1.0)
        Sample(sample={0: 1, 1: -1, 2: 1, 3: -1}, energy=-1.0)
        >>> # Snipped above response for brevity
        >>> # Unembed
        >>> source_response = dimod.Response.from_dicts(dimod.iter_unembed(target_response,
        ...                          embedding),
        ...                          {'energy': target_response.data_vectors['energy']})
        >>> for datum in source_response.data():    # doctest: +SKIP
        ...     print(datum)
        ...
        Sample(sample={'a': 1, 'c': -1, 'b': 1}, energy=-1.0)
        Sample(sample={'a': -1, 'c': -1, 'b': 1}, energy=-1.0)
        Sample(sample={'a': 1, 'c': 1, 'b': -1}, energy=-1.0)
        Sample(sample={'a': -1, 'c': 1, 'b': 1}, energy=-1.0)
        >>> # Snipped above response for brevity

    .. _dwave-system: https://github.com/dwavesystems/dwave-system

    """
    if chain_break_method is None:
        chain_break_method = majority_vote

    for target_sample in target_samples:
        for source_sample in chain_break_method(target_sample, embedding):
            yield source_sample


def unembed_response(target_response, embedding, source_bqm, chain_break_method=None):
    """Unembed the response.

    Construct a response for the source binary quadratic model (BQM) by unembedding the given
    response from the target BQM.

    Args:
        target_response (:obj:`.Response`):
            Response from the target BQM.

        embedding (dict):
            Mapping from source graph to target graph as a dict of form {s: {t, ...}, ...},
            where s is a source-model variable and t is a target-model variable.

        source_bqm (:obj:`.BinaryQuadraticModel`):
            Source binary quadratic model.

        chain_break_method (function, optional, default=:func:`.majority_vote`):
            Method used to resolve chain breaks.

    Returns:
        :obj:`.Response`:
            Response for the source binary quadratic model.

    Examples:
        This example embeds a Boolean AND gate,
        :math:`x_3 \Leftrightarrow x_1 \wedge x_2`, in a square-structured
        graph and samples with dimod reference structured sampler,
        `StructureComposite`, using the dimod reference `ExactSolver` sampler.
        The gate is represented as penalty model
        :math:`x_1 x_2 - 2(x_1+x_2)x_3 +3x_3`, which is submitted to the
        sampler as QUBO  problem
        :math:`E(a_i, b_{i,j}; x_i) = 3x_3 + x_1x_2 - 2x_1x_3 - 2x_2x_3`.
        This QUBO represents a fully connected :math:`K_3` graph.
        Samples are unembedded by :func:`.unembed_response` and show that
        only valid states of the AND gate have zero energy (e.g., only input
        :math:`x_1 x_2=1,1` results in :math:`z=1`), while invalid states have
        higher energy.

        >>> import dimod
        >>> import networkx as nx
        >>> # Binary quadratic model for the AND gate
        >>> Q = {('x1', 'x2'): 1, ('x1', 'z'): -2, ('x2', 'z'): -2, ('z', 'z'): 3}
        >>> bqm = dimod.BinaryQuadraticModel.from_qubo(Q)
        >>> # Embed the BQM in a structured dimod sampler defined by a square graph
        >>> target_graph = nx.cycle_graph(4)
        >>> sampler = dimod.StructureComposite(dimod.ExactSolver(),
        ...                          list(target_graph.nodes), list(target_graph.edges))
        >>> embedding = {'x1': {0}, 'x2': {1}, 'z': {2, 3}}
        >>> target_Q = dimod.embed_qubo(Q, embedding, sampler.adjacency)
        >>> # Sample on the target graph
        >>> target_response = sampler.sample_qubo(target_Q)
        >>> # Unembed samples back to the problem graph
        >>> source_response = dimod.unembed_response(target_response, embedding, bqm)
        >>> # Verify correct representation of the AND gate (first automatically then manually)
        >>> for datum in source_response.data():
        ...     if (datum.sample['x1'] and datum.sample['x2']) == datum.sample['z']:
        ...         if datum.energy > 0:
        ...            print('Valid AND has high energy')
        ...
        ...     else:
        ...         if datum.energy == 0:
        ...             print('invalid AND has low energy')
        ...
        >>> for datum in source_response.data():     # doctest: +SKIP
        ...     print(datum)
        ...
        Sample(sample={'x2': 0, 'x1': 0, 'z': 0}, energy=0.0)
        Sample(sample={'x2': 0, 'x1': 1, 'z': 0}, energy=0.0)
        Sample(sample={'x2': 1, 'x1': 0, 'z': 0}, energy=0.0)
        Sample(sample={'x2': 1, 'x1': 1, 'z': 1}, energy=0.0)
        Sample(sample={'x2': 1, 'x1': 0, 'z': 0}, energy=0.0)
        Sample(sample={'x2': 0, 'x1': 1, 'z': 0}, energy=0.0)
        Sample(sample={'x2': 0, 'x1': 1, 'z': 0}, energy=0.0)
        Sample(sample={'x2': 0, 'x1': 0, 'z': 0}, energy=0.0)
        Sample(sample={'x2': 1, 'x1': 0, 'z': 0}, energy=0.0)
        Sample(sample={'x2': 0, 'x1': 0, 'z': 0}, energy=0.0)
        Sample(sample={'x2': 1, 'x1': 1, 'z': 0}, energy=1.0)
        Sample(sample={'x2': 0, 'x1': 1, 'z': 1}, energy=1.0)
        Sample(sample={'x2': 1, 'x1': 0, 'z': 1}, energy=1.0)
        Sample(sample={'x2': 1, 'x1': 1, 'z': 0}, energy=1.0)
        Sample(sample={'x2': 1, 'x1': 1, 'z': 0}, energy=1.0)
        Sample(sample={'x2': 0, 'x1': 0, 'z': 1}, energy=3.0)

    """
    if any(v not in embedding for v in source_bqm):
        raise ValueError("given bqm does not match the embedding")

    energies = []

    def _samples():
        # populate energies as the samples are resolved one-at-a-time
        for sample in iter_unembed(target_response, embedding, chain_break_method):
            energies.append(source_bqm.energy(sample))
            yield sample

    # overwrite energy with the new values
    data_vectors = target_response.data_vectors.copy()
    data_vectors['energy'] = energies

    # NB: this works because response.from_dict does not resolve the energies until AFTER the samples
    return target_response.from_dicts(_samples(), data_vectors,
                                      vartype=target_response.vartype,
                                      info=target_response.info)



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
# ================================================================================================

from __future__ import absolute_import

import json
from pkg_resources import resource_filename

from six import iteritems

from dimod.binary_quadratic_model import BinaryQuadraticModel
from dimod.package_info import __version__
from dimod.response import Response
from dimod.vartypes import Vartype

bqm_json_schema_version = "1.0.0"

with open(resource_filename(__name__, 'bqm_json_schema.json'), 'r') as schema_file:
    bqm_json_schema = json.load(schema_file)


def bqm_decode_hook(dct):
    if 'bias' in dct:
        bias = dct['bias']
        if 'label' in dct:
            u = dct['label']
            if isinstance(u, list):
                u = tuple(u)
            return (u, bias)
        else:
            u = dct['label_head']
            v = dct['label_tail']
            if isinstance(u, list):
                u = tuple(u)
            if isinstance(v, list):
                v = tuple(v)
            return ((u, v), bias)

    elif 'linear_terms' in dct and 'quadratic_terms' in dct:
        return BinaryQuadraticModel(dict(dct['linear_terms']),
                                    dict(dct['quadratic_terms']),
                                    dct['offset'],
                                    Vartype[dct['variable_type']],
                                    **dct['info'])

    return dct


class DimodEncoder(json.JSONEncoder):
    """Subclass the JSONEncoder for dimod objects."""
    def default(self, obj):
        if isinstance(obj, BinaryQuadraticModel):

            if obj.vartype is Vartype.SPIN:
                vartype_string = 'SPIN'
            elif obj.vartype is Vartype.BINARY:
                vartype_string = 'BINARY'
            else:
                raise RuntimeError("unknown vartype")

            # by using the stream objects, we don't need to create the entire json_dict in memory
            json_dict = {"linear_terms": _JSONLinearBiasStream.from_dict(obj.linear),
                         "quadratic_terms": _JSONQuadraticBiasStream.from_dict(obj.quadratic),
                         "offset": obj.offset,
                         "variable_type": vartype_string,
                         "version": {"dimod": __version__, "bqm_schema": bqm_json_schema_version},
                         "variable_labels": _JSONLabelsStream.from_dict(obj.linear),
                         "info": obj.info}
            return json_dict

        if isinstance(obj, Response):
            # we will eventually want to implement this
            raise NotImplementedError

        return json.JSONEncoder.default(self, obj)


class _JSONLinearBiasStream(list):
    """Allows json to encode linear biases without needing to create a large number of dicts in
    memory.
    """
    @classmethod
    def from_dict(cls, linear):
        jlbs = cls()
        jlbs.linear = linear
        return jlbs

    def __iter__(self):
        for v, bias in iteritems(self.linear):
            yield {"bias": bias, "label": v}

    def __len__(self):
        return len(self.linear)


class _JSONQuadraticBiasStream(list):
    """Allows json to encode quadratic biases without needing to create a large number of dicts in
    memory.
    """
    @classmethod
    def from_dict(cls, quadratic):
        jqbs = cls()
        jqbs.quadratic = quadratic
        return jqbs

    def __iter__(self):
        for (u, v), bias in iteritems(self.quadratic):
            yield {"bias": bias, "label_head": u, "label_tail": v}

    def __len__(self):
        return len(self.quadratic)


class _JSONLabelsStream(list):
    """Allows json to encode variable labels without needing to create a large list in memory.
    """
    @classmethod
    def from_dict(cls, linear):
        jlbs = cls()
        jlbs.linear = linear
        return jlbs

    def __iter__(self):
        for u in self.linear:
            yield u

    def __len__(self):
        return len(self.linear)



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
# ================================================================================================
"""
dimod samplers respond with a consistent :class:`.Response` object that is
:class:`~collections.Iterable` (over samples, from lowest energy to highest) and
:class:`~collections.Sized` (number of samples).

Examples
--------
    This example shows the response of the dimod ExactSolver sampler.

>>> import dimod
>>> response = dimod.ExactSolver().sample_ising({'a': -0.5}, {})
>>> len(response)
2
>>> for sample in response:
...     print(sample)
{'a': 1}
{'a': -1}

"""
from collections import Mapping, Iterable, Sized, namedtuple, ValuesView, ItemsView
import itertools
import concurrent.futures

import numpy as np
from six import itervalues, iteritems

from dimod.decorators import vartype_argument
from dimod.utilities import resolve_label_conflict
from dimod.vartypes import Vartype

__all__ = ['Response']


class Response(Iterable, Sized):
    """A container for samples and any other data returned by dimod samplers.

    Args:
        samples_matrix (:obj:`numpy.matrix`):
            Samples as a NumPy matrix where each row is a sample.

        data_vectors (dict[field, :obj:`numpy.array`/list]):
            Additional per-sample data as a dict of vectors. Each vector is of the
            same length as `samples_matrix`. The key 'energy' and its vector are required.

        vartype (:class:`.Vartype`):
            Vartype of the samples.

        info (dict, optional, default=None):
            Information about the response as a whole formatted as a dict.

        variable_labels (list, optional, default=None):
            Variable labels mappped by index to columns of the samples matrix.

    Attributes:
        vartype (:class:`.Vartype`): Vartype of the samples.

        info (dict): Information about the response as a whole formatted as a dict.

        variable_labels (list/None): Variable labels. Each column in the samples matrix is the
            values returned for one variable. If None, column indices are the labels.

        label_to_idx (dict): Map of variable labels to columns in samples matrix.

    Examples:
        This example shows some attributes of the response for the sampler
        of dimod package's random_sampler.py reference example.

        >>> from dimod.reference.samplers.random_sampler import RandomSampler
        >>> sampler = RandomSampler()
        >>> bqm = dimod.BinaryQuadraticModel({0: 0.0, 1: 1.0}, {(0, 1): 0.5}, -0.5, dimod.SPIN)
        >>> response = sampler.sample(bqm)
        >>> response.vartype  # doctest: +SKIP
        <Vartype.SPIN: frozenset([1, -1])>  # doctest: +SKIP
        >>> response.variable_labels
        [0, 1]

    """

    @vartype_argument('vartype')
    def __init__(self, samples_matrix, data_vectors, vartype, info=None, variable_labels=None):
        # Constructor is opinionated about the samples_matrix type, it should be a numpy matrix
        if not isinstance(samples_matrix, np.matrix):
            raise TypeError("expected 'samples_matrix' to be a NumPy matrix")
        elif samples_matrix.dtype != np.int8:
            # cast to int8
            samples_matrix = samples_matrix.astype(np.int8)

        self._samples_matrix = samples_matrix
        num_samples, num_variables = samples_matrix.shape

        if not isinstance(data_vectors, Mapping):
            raise TypeError("expected 'data_vectors' to be a dict")
        if 'energy' not in data_vectors:
            raise ValueError("energy must be provided")
        else:
            data_vectors = dict(data_vectors)  # shallow copy

        for key, vector in iteritems(data_vectors):
            try:
                data_vectors[key] = vector = np.asarray(vector)
            except (ValueError, TypeError):
                raise TypeError("expected data vector {} to be array-like".format(key))

            shape = vector.shape
            if not shape or shape[0] != num_samples:
                raise ValueError(("expected data vector {} (shape {}) to have {} rows"
                                  "").format(key, vector.shape, num_samples))

        self._data_vectors = data_vectors

        # vartype is checked by the decorator
        self.vartype = vartype

        if info is None:
            info = {}
        elif not isinstance(info, dict):
            raise TypeError("expected 'info' to be a dict.")
        else:
            info = dict(info)  # make a shallow copy
        self.info = info

        if variable_labels is None:
            self.variable_labels = None
            self.label_to_idx = None
        else:
            self.variable_labels = variable_labels = list(variable_labels)
            if len(variable_labels) != num_variables:
                msg = ("variable_labels' length must match the number of columns in "
                       "samples_matrix, {} labels, matrix has {} columns".format(len(variable_labels), num_variables))
                raise ValueError(msg)

            self.label_to_idx = {v: idx for idx, v in enumerate(variable_labels)}

        # will store any pending Future objects and data about them
        self._futures = {}

    def __len__(self):
        """The number of samples."""
        num_samples, num_variables = self.samples_matrix.shape
        return num_samples

    def __iter__(self):
        """Iterate over the samples, low energy to high."""
        return self.samples(sorted_by='energy')

    def __str__(self):
        # developer note: it would be nice if the variable labels (if present could be printed)
        return self.samples_matrix.__str__()

    ##############################################################################################
    # Properties
    ##############################################################################################

    @property
    def samples_matrix(self):
        """:obj:`numpy.matrix`: Samples as a NumPy matrix of data type int8.

        Examples:
            This example shows the samples of dimod package's ExactSolver reference sampler
            formatted as a NumPy matrix.

            >>> import dimod
            >>> response = dimod.ExactSolver().sample_ising({'a': -0.5, 'b': 1.0}, {('a', 'b'): -1})
            >>> response.samples_matrix
            matrix([[-1, -1],
                    [ 1, -1],
                    [ 1,  1],
                    [-1,  1]])

        """
        if self._futures:
            self._resolve_futures(**self._futures)

        return self._samples_matrix

    @samples_matrix.setter
    def samples_matrix(self, mat):
        self._samples_matrix = mat

    @property
    def data_vectors(self):
        """dict[field, :obj:`numpy.array`/list]: Per-sample data as a dict, where keys are the
        data labels and values are each a vector of the same length as sample_matrix.

        Examples:
            This example shows the returned energies of dimod package's ExactSolver
            reference sampler.

            >>> import dimod
            >>> response = dimod.ExactSolver().sample_ising({'a': -0.5, 'b': 1.0}, {('a', 'b'): -1})
            >>> response.data_vectors['energy']
            array([-1.5, -0.5, -0.5,  2.5])

        """
        if self._futures:
            self._resolve_futures(**self._futures)

        return self._data_vectors

    def done(self):
        """True if all loaded futures are done or if there are no futures.

        Only relevant when the response is constructed with :meth:`Response.from_futures`.

        Examples:
            This example checks whether futures are done before and after a `set_result` call.

            >>> from concurrent.futures import Future
            >>> future = Future()
            >>> response = dimod.Response.from_futures((future,), dimod.BINARY, 3)
            >>> future.done()
            False
            >>> future.set_result({'samples': [0, 1, 0], 'energy': [1]})
            >>> future.done()
            True

        """
        return all(future.done() for future in self._futures.get('futures', tuple()))

    ##############################################################################################
    # Construction and updates
    ##############################################################################################

    @classmethod
    def from_matrix(cls, samples, data_vectors, vartype=None, info=None, variable_labels=None):
        """Build a response from a NumPy array-like object.

        Args:
            samples (array_like/str):
                Samples as a :class:`numpy.matrix` or NumPy array-like object. See Notes.

            data_vectors (dict[field, :obj:`numpy.array`/list]):
                Additional per-sample data as a dict of vectors. Each vector is the same length as
                `samples_matrix`. The key 'energy' and its vector are required.

            vartype (:class:`.Vartype`, optional, default=None):
                Vartype of the response. If not provided, vartype is inferred from the
                samples matrix if possible or a ValueError is raised.

            info (dict, optional, default=None):
                Information about the response as a whole formatted as a dict.

            variable_labels (list, optional, default=None):
                Maps (by index) variable labels to the columns of the samples matrix.

        Returns:
            :obj:`.Response`: A `dimod` :obj:`.Response` object based on the input
            NumPy array-like object.

        Raises:
            :exc:`ValueError`: If vartype is not provided and samples are all 1s, have more
                than two unique values, or have values of an unknown vartype.

        Examples:
            This example code snippet builds a response from a NumPy matrix.

            .. code-block:: python

                samples = np.matrix([[0, 1], [1, 0]])
                energies = [0.0, 1.0]
                response = Response.from_matrix(samples, {'energy': energies})

            This example code snippet builds a response from a NumPy array-like object (a Python list).

            .. code-block:: python

                samples = [[0, 1], [1, 0]]
                energies = [0.0, 1.0]
                response = Response.from_matrix(samples, {'energy': energies})

        Notes:
            SciPy defines array_like in the following way: "In general, numerical data arranged in
            an array-like structure in Python can be converted to arrays through the use of the
            array() function. The most obvious examples are lists and tuples. See the documentation
            for array() for details for its use. Some objects may support the array-protocol and
            allow conversion to arrays this way. A simple way to find out if the object can be
            converted to a numpy array using array() is simply to try it interactively and see if it
            works! (The Python Way)." [array_like]_

        References:
            .. [array_like] Docs.scipy.org. (2018). Array creation - NumPy v1.14 Manual. [online]
                Available at: https://docs.scipy.org/doc/numpy/user/basics.creation.html
                [Accessed 16 Feb. 2018].

        """
        samples_matrix = np.matrix(samples, dtype=np.int8)

        if vartype is None:
            vartype = infer_vartype(samples_matrix)

        response = cls(samples_matrix, data_vectors=data_vectors,
                       vartype=vartype, info=info, variable_labels=variable_labels)

        return response

    @classmethod
    def from_dicts(cls, samples, data_vectors, vartype=None, info=None):
        """Build a response from an iterable of dicts.

        Args:
            samples (iterable[dict]):
                Iterable of samples where each sample is a dictionary (or Mapping).

            data_vectors (dict[field, :obj:`numpy.array`/list]):
                Additional per-sample data as a dict of vectors. Each vector is the
                same length as `samples_matrix`. The key 'energy' and its vector are required.

            vartype (:class:`.Vartype`, optional, default=None):
                Vartype of the response. If not provided, vartype is inferred from the
                samples matrix if possible or a ValueError is raised.

            info (dict, optional, default=None):
                Information about the response as a whole formatted as a dict.

        Returns:
            :obj:`.Response`: A `dimod` :obj:`.Response` object based on the input dicts.

        Raises:
            :exc:`ValueError`: If vartype is not provided and samples are all 1s, have more
                than two unique values, or have values of an unknown vartype.

        Examples:
            This example code snippet builds a response from an interable of samples and
            its corresponding dict of energies.

            .. code-block:: python

                samples = [{'a': -1, 'b': +1}, {'a': +1, 'b': -1}]
                energies = [-1.0, -1.0]
                response = Response.from_dicts(samples, {'energy': energies})

        """
        samples = iter(samples)

        # get the first sample
        first_sample = next(samples)

        try:
            variable_labels = sorted(first_sample)
        except TypeError:
            # unlike types cannot be sorted in python3
            variable_labels = list(first_sample)
        num_variables = len(variable_labels)

        def _iter_samples():
            yield np.fromiter((first_sample[v] for v in variable_labels),
                              count=num_variables, dtype=np.int8)

            try:
                for sample in samples:
                    yield np.fromiter((sample[v] for v in variable_labels),
                                      count=num_variables, dtype=np.int8)
            except KeyError:
                msg = ("Each dict in 'samples' must have the same keys.")
                raise ValueError(msg)

        samples_matrix = np.matrix(np.stack(list(_iter_samples())))

        return cls.from_matrix(samples_matrix, data_vectors=data_vectors, vartype=vartype,
                               info=info, variable_labels=variable_labels)

    @classmethod
    def from_pandas(cls, samples_df, data_vectors, vartype=None, info=None):
        """Build a response from a pandas DataFrame.

        Args:
            samples (:obj:`pandas.DataFrame`):
                A pandas DataFrame of samples where each row is a sample.

            data_vectors (dict[field, :obj:`numpy.array`/list]):
                Additional per-sample data as a dict of vectors. Each vector is the
                same length as `samples_matrix`. The key 'energy' and its vector are required.

            vartype (:class:`.Vartype`, optional, default=None):
                Vartype of the response. If not provided, vartype is inferred from the
                samples matrix if possible or a ValueError is raised.

            info (dict, optional, default=None):
                Information about the response as a whole formatted as a dict.

        Returns:
            :obj:`.Response`: A `dimod` :obj:`.Response` object based on the input DataFrame.

        Raises:
            :exc:`ValueError`: If vartype is not provided and samples are all 1s, have more
                than two unique values, or have values of an unknown vartype.

        Examples:
            These example code snippets build a response from a pandas DataFrame.

            .. code-block:: python

                import pandas as pd

                samples = pd.DataFrame([{'a': 1, 'b': 0}, {'a': 0, 'b': 0}], dtype='int8')
                response = Response.from_pandas(samples, {energy: [1, 0]})

            .. code-block:: python

                import pandas as pd

                samples = pd.DataFrame([[+1, -1]], dtype='int8', columns=['v1', 'v2'])
                response = Response.from_pandas(samples, {energy: [1]})

        """
        import pandas as pd

        variable_labels = list(samples_df.columns)
        samples_matrix = np.matrix(samples_df.values)

        if isinstance(data_vectors, pd.DataFrame):
            raise NotImplementedError("support for DataFrame data_vectors is forthcoming")

        return cls.from_matrix(samples_matrix, data_vectors, vartype=vartype, info=info,
                               variable_labels=variable_labels)

    @classmethod
    def empty(cls, vartype):
        empty_samples_matrix = np.matrix(np.empty((0, 0), dtype=np.int8))
        empty_data_vectors = {'energy': []}
        return cls(samples_matrix=empty_samples_matrix, data_vectors=empty_data_vectors,
                   vartype=vartype)

    @classmethod
    def from_futures(cls, futures, vartype, num_variables,
                     samples_key='samples', data_vector_keys=None,
                     info_keys=None, variable_labels=None, active_variables=None,
                     ignore_extra_keys=True):
        """Build a response from :obj:`~concurrent.futures.Future`-like objects.

        Args:
            futures (iterable):
                Iterable :obj:`~concurrent.futures.Future` or :obj:`~concurrent.futures.Future`-like
                objects (Python objects with similar structure).
                :meth:`~concurrent.futures.Future.result` returns a dict.

            vartype (:class:`.Vartype`):
                Vartype of the response.

            num_variables (int):
                Number of variables for each sample.

            samples_key (hashable, optional, default='samples'):
                Key of the result dict containing the samples. Samples are array-like.

            data_vector_keys (iterable/mapping, optional, default=None):
                A mapping from the keys of the result dict to :attr:`Response.data_vectors`. If
                None, ['energy'] is assumed to be a key in the result dict and the
                'energy' data vector is mapped.

            info_keys (iterable/mapping, optional, default=None):
                A mapping from the keys of the result dict to :attr:`Response.info`.
                If None, info is empty.

            variable_labels (list, optional, default=None):
                Maps (by index) variable labels to columns of the samples matrix.

            active_variables (array-like, optional, default=None):
                Selects which columns of the result's samples are used. If `variable_labels` is
                not provided, `variable_labels` is set to match `active_variables`.

            ignore_extra_keys (bool, optional, default=True):
                If True, keys given in `data_vector_keys` and `info_keys` but that are not in
                :meth:`~concurrent.futures.Future.result` are ignored. If False, extra keys
                cause a ValueError.

        Returns:
            :obj:`.Response`: A `dimod` :obj:`.Response` object based on the input
            Future-like objects.

        Notes:
            :obj:`~concurrent.futures.Future` objects are read on the first read
            of :attr:`.Response.samples_matrix` or :attr:`.Response.data_vectors`.

        Examples:
            These example code snippets build responses from :obj:`~concurrent.futures.Future` objects.

            .. code-block:: python

                from concurrent.futures import Future

                future = Future()

                # load the future into response
                response = dimod.Response.from_futures((future,), dimod.BINARY, 3)

                future.set_result({'samples': [0, 1, 0], 'energy': [1]})

                # now read from the response
                matrix = response.samples_matrix

            .. code-block:: python

                from concurrent.futures import Future

                future = Future()

                # load the future into response
                response = dimod.Response.from_futures((future,), dimod.BINARY, 3,
                                                       active_variables=[0, 1, 3])

                future.set_result({'samples': [0, 1, 3, 0], 'energy': [1]})

                # now read from the response, this matrix
                matrix = response.samples_matrix

            .. code-block:: python

                from concurrent.futures import Future

                future = Future()

                # load the future into response
                response = dimod.Response.from_futures((future,), dimod.BINARY, 3,
                                                       data_vector_keys={'en': 'energy'})

                future.set_result({'samples': [0, 1, 0], 'en': [1]})

                # now read from the response
                matrix = response.samples_matrix

        """

        if data_vector_keys is None:
            data_vector_keys = {'energy': 'energy'}
        elif isinstance(data_vector_keys, Mapping):
            data_vector_keys = dict(data_vector_keys)
        else:
            data_vector_keys = {key: key for key in data_vector_keys}  # identity mapping

        if info_keys is None:
            info_keys = {}
        elif isinstance(info_keys, Mapping):
            info_keys = dict(info_keys)
        else:
            info_keys = {key: key for key in info_keys}

        if active_variables is not None:
            if variable_labels is None:
                variable_labels = active_variables
            elif len(variable_labels) != len(active_variables):
                raise ValueError("active_variables and variable_labels should have the same length")

        response = cls.empty(vartype)

        # now dump all of the remaining information into the _futures
        response._futures = {'futures': list(futures),
                             'samples_key': samples_key,
                             'data_vector_keys': data_vector_keys,
                             'info_keys': info_keys,
                             'variable_labels': variable_labels,
                             'active_variables': active_variables,
                             'ignore_extra_keys': ignore_extra_keys}

        return response

    def _resolve_futures(self, futures, samples_key, data_vector_keys, info_keys,
                         variable_labels, active_variables, ignore_extra_keys):

        # first reset the futures to avoid recursion errors
        self._futures = {}

        # `dwave.cloud.qpu.computation.Future` is not yet interchangeable with
        # `concurrent.futures.Future`, so we need to detect the kind of future
        # we're dealing with.
        futures = list(futures)  # if generator
        if hasattr(futures[0], 'as_completed'):
            as_completed = futures[0].as_completed
        else:
            as_completed = concurrent.futures.as_completed

        # combine all samples from all futures into a single response
        for future in as_completed(futures):
            result = dict(future.result())  # create a shallow copy

            # first get the samples matrix and filter out any inactive variables
            samples = np.matrix(result.pop(samples_key), dtype=np.int8)
            if active_variables is not None:
                samples = samples[:, active_variables]

            # next get the data vectors
            if ignore_extra_keys:
                data_vectors = {}
                for source_key, key in iteritems(data_vector_keys):
                    try:
                        data_vectors[key] = result[source_key]
                    except KeyError:
                        pass
                info = {}
                for source_key, key in iteritems(info_keys):
                    try:
                        info[key] = result[source_key]
                    except KeyError:
                        pass
            else:
                data_vectors = {}
                for source_key, key in iteritems(data_vector_keys):
                    try:
                        data_vectors[key] = result[source_key]
                    except KeyError:
                        raise ValueError("data vector key '{}' not in Future.result()".format(key))
                info = {}
                for source_key, key in iteritems(info_keys):
                    try:
                        info[key] = result[source_key]
                    except KeyError:
                        raise ValueError("info key '{}' not in Future.result()".format(key))

            # now get the appropriate response
            response = self.__class__.from_matrix(samples, data_vectors=data_vectors, info=info,
                                                  vartype=self.vartype, variable_labels=variable_labels)
            self.update(response)

    def update(self, *other_responses):
        """Add values of other responses to the response.

        Args:
            *other_responses: (:obj:`.Response`):
                Additional responses from which to add values. Any number of additional response objects,
                separated by commas, can be specified. Responses must have matching `sample_matrix`
                dimensions, `data_vector` keys, and variable labels.

        Examples:
            This example updates a response with values from two other responses.

            >>> import dimod
            >>> samples = [[0, 1], [1, 0]]
            >>> energies = [0.0, 1.0]
            >>> response = dimod.Response.from_matrix(samples, {'energy': energies})
            >>> samples1 = [[0, 0], [1, 1]]
            >>> energies1 = [0.25, 1.25]
            >>> response1 = dimod.Response.from_matrix(samples1, {'energy': energies1})
            >>> samples2 = [[1, 0], [0, 1]]
            >>> energies2 = [0.5, 1.75]
            >>> response2 = dimod.Response.from_matrix(samples2, {'energy': energies2})
            >>> len(response)
            2
            >>> for i in response.data():         # doctest: +SKIP
            ...     print(i)
            ...
            Sample(sample={0: 0, 1: 1}, energy=0.0)
            Sample(sample={0: 1, 1: 0}, energy=1.0)
            >>> response.update(response1, response2)
            >>> len(response)
            6
            >>> for energy in response.data(fields=['energy'], name='UpdatedEnergy'):
            ...     print(energy)
            ...
            UpdatedEnergy(energy=0.0)
            UpdatedEnergy(energy=0.25)
            UpdatedEnergy(energy=0.5)
            UpdatedEnergy(energy=1.0)
            UpdatedEnergy(energy=1.25)
            UpdatedEnergy(energy=1.75)

        """
        # make sure all of the other responses are the appropriate vartype. We could cast them but
        # that would effect the energies so it is best to happen outside of this function.
        vartype = self.vartype
        for response in other_responses:
            if vartype is not response.vartype:
                raise ValueError("can only update with responses of matching vartype base")

        # if self is empty, then we are done
        if not self:
            other_responses = list(other_responses)

            response = other_responses.pop()

            self.samples_matrix = response.samples_matrix
            self.data_vectors.update(response.data_vectors)
            self.info.update(response.info)
            self.variable_labels = response.variable_labels
            self.label_to_idx = response.label_to_idx

            if other_responses:
                self.update(*other_responses)

            return

        # make sure that the variable labels are consistent
        variable_labels = self.variable_labels
        if variable_labels is None:
            __, num_variables = self.samples_matrix.shape
            variable_labels = list(range(num_variables))
            # in this case we need to allow for either None or variable_labels
            if not all(response.variable_labels is None or response.variable_labels == variable_labels
                       for response in other_responses):
                raise ValueError("cannot update responses with unlike variable labels")
        else:
            if not all(response.variable_labels == variable_labels for response in other_responses):
                raise ValueError("cannot update responses with unlike variable labels")

        # concatenate all of the matrices
        matrices = [self.samples_matrix]
        matrices.extend([response.samples_matrix for response in other_responses])
        self.samples_matrix = np.concatenate(matrices)

        # group all of the data vectors
        for key in self.data_vectors:
            vectors = [self.data_vectors[key]]
            vectors.extend(response.data_vectors[key] for response in other_responses)
            self.data_vectors[key] = np.concatenate(vectors)

        # finally update the response info
        for response in other_responses:
            self.info.update(response.info)

    ###############################################################################################
    # Transformations and Copies
    ###############################################################################################

    def copy(self):
        """Creates a shallow copy of a response.

        Examples:
            This example copies a response.

            >>> import dimod
            >>> samples = [[0, 1], [1, 0]]
            >>> energies = [0.0, 1.0]
            >>> response = dimod.Response.from_matrix(samples, {'energy': energies})
            >>> copied_response = response.copy()

        """
        return self.from_matrix(self.samples_matrix, self.data_vectors,
                                vartype=self.vartype, info=self.info,
                                variable_labels=self.variable_labels)

    @vartype_argument('vartype')
    def change_vartype(self, vartype, data_vector_offsets=None, inplace=True):
        """Create a new response with the given vartype.

        Args:
            vartype (:class:`.Vartype`/str/set):
                Variable type to use for the new response. Accepted input values:

                * :class:`.Vartype.SPIN`, ``'SPIN'``, ``{-1, 1}``
                * :class:`.Vartype.BINARY`, ``'BINARY'``, ``{0, 1}``

            data_vector_offsets (dict[field, :obj:`numpy.array`/list], optional, default=None):
                Offsets to add to `data_vectors` of the response formatted as a dict containing
                per-sample offsets in vectors. Each vector is the same length as `samples_matrix`.

            inplace (bool, optional, default=True):
                If True, the response is updated in-place, otherwise a new response is returned.

        Returns:
            :obj:`.Response`. New response with vartype matching input 'vartype'.

        Examples:
            This example converts the response of the dimod package's ExactSolver sampler
            to binary and adds offsets.

            >>> import dimod
            >>> response = dimod.ExactSolver().sample_ising({'a': -0.5, 'b': 1.0}, {('a', 'b'): -1})
            >>> response_binary = response.change_vartype('BINARY',
            ...                   data_vector_offsets={'energy': [0, 0.1, 0.2, 0.3]},
            ...                   inplace=False)
            >>> response_binary.vartype
            <Vartype.BINARY: frozenset([0, 1])>
            >>> for datum in response_binary.data():    # doctest: +SKIP
            ...    print(datum)
            ...
            Sample(sample={'a': 0, 'b': 0}, energy=-1.5)
            Sample(sample={'a': 1, 'b': 0}, energy=-0.4)
            Sample(sample={'a': 1, 'b': 1}, energy=-0.3)
            Sample(sample={'a': 0, 'b': 1}, energy=2.8)

            This example code snippet creates a response with spin variables from a response
            with binary variables while adding energy offsets to the new response.

            .. code-block:: python

                import pandas as pd
                samples = [[0, 1], [1, 0]]
                energies = [0.0, 1.0]
                response = dimod.Response.from_matrix(samples, {'energy': energies})
                offsets = {'energy': [0.25, 0.75]}
                response.change_vartype('SPIN',
                                         data_vector_offsets = offsets)
        """
        if not inplace:
            return self.copy().change_vartype(vartype, data_vector_offsets=data_vector_offsets, inplace=True)

        if data_vector_offsets is not None:
            for key in data_vector_offsets:
                self.data_vectors[key] += data_vector_offsets[key]

        if vartype is self.vartype:
            return self

        if vartype is Vartype.SPIN and self.vartype is Vartype.BINARY:
            self.samples_matrix = 2 * self.samples_matrix - 1
            self.vartype = vartype
        elif vartype is Vartype.BINARY and self.vartype is Vartype.SPIN:
            self.samples_matrix = (self.samples_matrix + 1) // 2
            self.vartype = vartype
        else:
            raise ValueError("Cannot convert from {} to {}".format(self.vartype, vartype))

        return self

    def relabel_variables(self, mapping, inplace=True):
        """Relabel a response's variables as per a given mapping.

        Args:
            mapping (dict):
                Dict mapping current variable labels to new. If an incomplete mapping is
                provided, unmapped variables keep their original labels

            inplace (bool, optional, default=True):
                If True, the original response is updated; otherwise a new response is returned.

        Returns:
            :class:`.Response`: Response with relabeled variables. If inplace=True, returns
            itself.

        Examples:
            This example relabels the response of the dimod package's ExactSolver sampler and
            saves it as a new response.

            >>> import dimod
            >>> response = dimod.ExactSolver().sample_ising({'a': -0.5, 'b': 1.0}, {('a', 'b'): -1})
            >>> new_response = response.relabel_variables({'a': 0, 'b': 1}, inplace=False)
            >>> [next(new_response.samples())[x] for x in [0, 1]]
            [-1, -1]            


            This example code snippet relabels variables in a response.

            .. code-block:: python

                response = dimod.Response.from_dicts([{'a': -1}, {'a': +1}], {'energy': [-1, 1]})
                response.relabel_variables({'a': 0})

        """
        if not inplace:
            return self.copy().relabel_variables(mapping, inplace=True)

        # we need labels
        if self.variable_labels is None:
            __, num_variables = self.samples_matrix.shape
            self.variable_labels = list(range(num_variables))

        try:
            old_labels = set(mapping)
            new_labels = set(itervalues(mapping))
        except TypeError:
            raise ValueError("mapping targets must be hashable objects")

        for v in new_labels:
            if v in self.variable_labels and v not in old_labels:
                raise ValueError(('A variable cannot be relabeled "{}" without also relabeling '
                                  "the existing variable of the same name").format(v))

        shared = old_labels & new_labels
        if shared:
            old_to_intermediate, intermediate_to_new = resolve_label_conflict(mapping, old_labels, new_labels)

            self.relabel_variables(old_to_intermediate, inplace=True)
            self.relabel_variables(intermediate_to_new, inplace=True)
            return self

        self.variable_labels = variable_labels = [mapping.get(v, v) for v in self.variable_labels]
        self.label_to_idx = {v: idx for idx, v in enumerate(variable_labels)}
        return self

    ###############################################################################################
    # Viewing a Response
    ###############################################################################################

    def samples(self, n=None, sorted_by='energy'):
        """Iterate over the samples in the response.

        Args:
            n (int, optional, default=None):
                The maximum number of samples to provide. If None, all are provided.

            sorted_by (str/None, optional, default='energy'):
                Selects the `data_vector` used to sort the samples. If None, the samples are yielded in
                the order given by the samples matrix.

        Yields:
            :obj:`.SampleView`: A view object mapping the variable labels to their values. Acts like
            a read-only dict.

        Examples:
            This example iterates over the response samples of the dimod ExactSolver sampler.

            >>> import dimod
            >>> response = dimod.ExactSolver().sample_ising({'a': -0.5, 'b': 1.0}, {('a', 'b'): -1})
            >>> for sample in response.samples():    # sorted_by='energy'
            ...     print(sample['a']==sample['b'])
            ...
            True
            False
            True
            False
            >>> for sample in response.samples(sorted_by=None):   # doctest: +SKIP
            ...     print(sample)
            ...
            {'a': -1, 'b': -1}
            {'a': 1, 'b': -1}
            {'a': 1, 'b': 1}
            {'a': -1, 'b': 1}

        """
        num_samples = len(self)

        if n is not None:
            for sample in itertools.islice(self.samples(n=None, sorted_by=sorted_by), n):
                yield sample
            return

        if sorted_by is None:
            order = np.arange(num_samples)
        else:
            order = np.argsort(self.data_vectors[sorted_by])

        samples = self.samples_matrix
        label_mapping = self.label_to_idx

        for idx in order:
            yield SampleView(idx, self)

    def data(self, fields=None, sorted_by='energy', name='Sample'):
        """Iterate over the data in the response.

        Args:
            fields (list, optional, default=None):
                If specified, only these fields' values are included in the yielded tuples.
                The special field name 'sample' can be used to view the samples.

            sorted_by (str/None, optional, default='energy'):
                Selects the data_vector used to sort the samples. If None, the samples are yielded
                in the order given by the samples matrix.

            name (str/None, optional, default='Sample'):
                Name of the yielded namedtuples or None to yield regular tuples.

        Yields:
            namedtuple/tuple: The data in the response, in the order specified by the input
            `fields`.

        Examples:
            This example iterates over the response data of the dimod ExactSolver sampler.

            >>> import dimod
            >>> response = dimod.ExactSolver().sample_ising({'a': -0.5, 'b': 1.0}, {('a', 'b'): -1})
            >>> for datum in response.data():   # doctest: +SKIP
            ...     print(datum)
            ...
            Sample(sample={'a': -1, 'b': -1}, energy=-1.5)
            Sample(sample={'a': 1, 'b': -1}, energy=-0.5)
            Sample(sample={'a': 1, 'b': 1}, energy=-0.5)
            Sample(sample={'a': -1, 'b': 1}, energy=2.5)
            >>> for energy, in response.data(fields=['energy'], sorted_by='energy'):
            ...     print(energy)
            ...
            -1.5
            -0.5
            -0.5
            2.5
            >>> print(next(response.data(fields=['energy'], name='ExactSolverSample')))
            ExactSolverSample(energy=-1.5)

        """
        if fields is None:
            fields = ['sample']
            fields.extend(self.data_vectors)

        if sorted_by is None:
            order = np.arange(len(self))
        else:
            order = np.argsort(self.data_vectors[sorted_by])

        if name is None:
            # yielding a tuple
            def _pack(values):
                return tuple(values)
        else:
            # yielding a named tuple
            SampleTuple = namedtuple(name, fields)

            def _pack(values):
                return SampleTuple(*values)

        samples = self.samples_matrix
        label_mapping = self.label_to_idx
        data_vectors = self.data_vectors

        def _values(idx):
            for field in fields:
                if field == 'sample':
                    yield SampleView(idx, self)
                else:
                    yield data_vectors[field][idx]

        for idx in order:
            yield _pack(_values(idx))


class SampleView(Mapping):
    """View each row of the samples matrix as if it was a dict."""
    def __init__(self, idx, response):
        self._idx = idx  # row of response._samples_matrix
        self._response = response

    def __getitem__(self, key):
        label_mapping = self._response.label_to_idx
        if label_mapping is not None:
            key = label_mapping[key]
        return int(self._response.samples_matrix[self._idx, key])

    def __iter__(self):
        # iterate over the variables
        label_mapping = self._response.label_to_idx
        if label_mapping is None:
            __, num_variables = self._response.samples_matrix.shape
            return iter(range(num_variables))
        return label_mapping.__iter__()

    def __len__(self):
        __, num_variables = self._response.samples_matrix.shape
        return num_variables

    def __repr__(self):
        """Represents itself as as a dictionary"""
        return dict(self).__repr__()

    def values(self):
        return SampleValuesView(self)

    def items(self):
        return SampleItemsView(self)


class SampleItemsView(ItemsView):
    """Faster read access to the numpy matrix"""
    __slots__ = ()

    def __iter__(self):
        # Inherited __init__ puts the Mapping into self._mapping
        variable_labels = self._mapping._response.variable_labels
        samples_matrix = self._mapping._response.samples_matrix
        idx = self._mapping._idx
        if variable_labels is None:
            for v, val in enumerate(np.nditer(samples_matrix[idx, :], order='C', op_flags=['readonly'])):
                yield (v, int(val))
        else:
            for v, val in zip(variable_labels, np.nditer(samples_matrix[idx, :], order='C', op_flags=['readonly'])):
                yield (v, int(val))


class SampleValuesView(ValuesView):
    """Faster read access to the numpy matrix"""
    __slots__ = ()

    def __iter__(self):
        # Inherited __init__ puts the Mapping into self._mapping
        samples_matrix = self._mapping._response.samples_matrix
        for val in np.nditer(samples_matrix[self._mapping._idx, :], op_flags=['readonly']):
            yield int(val)


def infer_vartype(samples_matrix):
    """Try to determine the Vartype of the samples matrix based on its values.

    Args:
        samples_matrix (:object:`numpy.ndarray`):
            An array or matrix of samples.

    Returns:
        :class:`.Vartype`

    Raises:
        ValueError: If the matrix is all ones, contains values other than -1, 1, 0 or contains more
        than two unique values.

    """
    ones_matrix = samples_matrix == 1

    if np.all(ones_matrix):
        msg = ("ambiguous vartype - an empty samples_matrix or one where all the values "
               "are all 1 must have the vartype specified by setting vartype=dimod.SPIN or "
               "vartype=dimod.BINARY.")
        raise ValueError(msg)

    if np.all(ones_matrix + (samples_matrix == 0)):
        return Vartype.BINARY
    elif np.all(ones_matrix + (samples_matrix == -1)):
        return Vartype.SPIN
    else:
        sample_vals = set(int(v) for v in np.nditer(samples_matrix)) - {-1, 1, 0}
        if sample_vals:
            msg = ("samples_matrix includes unknown values {}").format(sample_vals)
        else:
            msg = ("samples_matrix includes both -1 and 0 values")
        raise ValueError(msg)



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
# ================================================================================================
"""
On the D-Wave system, coupling :math:`J_{i,j}` adds a small bias to qubits :math:`i` and
:math:`j` due to leakage. This can become significant for chained qubits. Additionally,
qubits are biased to some small degree in one direction or another.
Applying a spin-reversal transform can improve results by reducing the impact of possible
analog and systematic errors. A spin-reversal transform does not alter the Ising problem;
the transform simply amounts to reinterpreting spin up as spin down, and visa-versa, for
a particular spin.
"""
from random import random
import time
import itertools

from dimod.core.composite import Composite
from dimod.core.sampler import Sampler
from dimod.core.structured import Structured
from dimod.response import Response
from dimod.vartypes import Vartype

__all__ = ['SpinReversalTransformComposite']


class SpinReversalTransformComposite(Sampler, Composite):
    """Composite for applying spin reversal transform preprocessing.

    Spin reversal transforms (or "gauge transformations") are applied
    by flipping the spin of variables in the Ising problem. After
    sampling the transformed Ising problem, the same bits are flipped in the
    resulting sample.

    Args:
        sampler: A `dimod` sampler object.

    Attributes:
        children (list):
            [`sampler`] where `sampler` is the input sampler.
        structure:
            Inherited from input `sampler`.

    Examples:
        This example composes a dimod ExactSolver sampler with spin transforms then
        uses it to sample an Ising problem.

        >>> import dimod
        >>> # Compose the sampler
        >>> base_sampler = dimod.ExactSolver()
        >>> composed_sampler = dimod.SpinReversalTransformComposite(base_sampler)
        >>> base_sampler in composed_sampler.children
        True
        >>> # Sample an Ising problem
        >>> response = composed_sampler.sample_ising({'a': -0.5, 'b': 1.0}, {('a', 'b'): -1})
        >>> print(next(response.data()))           # doctest: +SKIP
        Sample(sample={'a': 1, 'b': 1}, energy=-1.5)

    References
    ----------
    .. [KM] Andrew D. King and Catherine C. McGeoch. Algorithm engineering
        for a quantum annealing platform. https://arxiv.org/abs/1410.2628,
        2014.

    """
    children = None
    parameters = None
    properties = None

    def __init__(self, child):
        self.children = [child]

        if isinstance(child, Structured):
            # todo something like Structured.__init__(self)
            raise NotImplementedError

        self.parameters = parameters = {'num_spin_reversal_transforms': [],
                                        'spin_reversal_variables': []}
        parameters.update(child.parameters)

        self.properties = {'child_properties': child.properties}

    def sample(self, bqm, num_spin_reversal_transforms=2, spin_reversal_variables=None, **kwargs):
        """Sample from the binary quadratic model.

        Args:
            bqm (:obj:`~dimod.BinaryQuadraticModel`):
                Binary quadratic model to be sampled from.
            num_spin_reversal_transforms (integer, optional, default=2):
                Number of spin reversal transform runs.
            spin_reversal_variables (list/dict, optional, default=None):
                Variables to which to apply the spin reversal. If None, every variable
                has a 50% probability of being selected.

        Returns:
            :obj:`~dimod.Response`: A `dimod` :obj:`.~dimod.Response` object.

        Examples:
            This example runs 100 spin reversals applied to one variable of a QUBO problem.

            >>> import dimod
            >>> base_sampler = dimod.ExactSolver()
            >>> composed_sampler = dimod.SpinReversalTransformComposite(base_sampler)
            >>> Q = {('a', 'a'): -1, ('b', 'b'): -1, ('a', 'b'): 2}
            >>> response = composed_sampler.sample_qubo(Q,
            ...               num_spin_reversal_transforms=100,
            ...               spin_reversal_variables={'a'})
            >>> len(response)
            400
            >>> print(next(response.data()))           # doctest: +SKIP
            Sample(sample={'a': 0, 'b': 1}, energy=-1.0)

        """
        # make a main response
        response = None

        for ii in range(num_spin_reversal_transforms):
            if spin_reversal_variables is None:
                # apply spin transform to each variable with 50% chance
                transform = list(v for v in bqm.linear if random() > .5)
            else:
                transform = list(spin_reversal_variables)

            flipped_bqm = bqm.copy()

            for v in transform:
                flipped_bqm.flip_variable(v)

            flipped_response = self.child.sample(bqm, **kwargs)

            tf_idxs = [flipped_response.label_to_idx[v] for v in flipped_response.variable_labels]

            if bqm.vartype is Vartype.SPIN:
                flipped_response.samples_matrix[:, tf_idxs] = -1 * flipped_response.samples_matrix[:, tf_idxs]
            else:
                flipped_response.samples_matrix[:, tf_idxs] = 1 - flipped_response.samples_matrix[:, tf_idxs]

            if response is None:
                response = flipped_response
            else:
                response.update(flipped_response)

        return response



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
# ================================================================================================

from __future__ import absolute_import

import abc


class SamplerABCMeta(abc.ABCMeta):
    def __new__(mcls, name, bases, namespace, **kwargs):
        cls = abc.ABCMeta.__new__(mcls, name, bases, namespace, **kwargs)

        samplermixins = {name
                         for name, value in namespace.items()
                         if getattr(value, "__issamplemixin__", False)}
        if len(samplermixins) == 3:
            abstracts = samplermixins
        else:
            abstracts = set()

        for base in bases:
            samplermixins = {name
                             for name in getattr(base, "__abstractmethods__", set())
                             if getattr(getattr(cls, name, None), "__issamplemixin__", False)}
            if len(samplermixins) == 3:
                abstracts.update(samplermixins)

        # if we found any, update abstract methods
        if abstracts:
            cls.__abstractmethods__ = frozenset(abstracts.union(cls.__abstractmethods__))

        return cls


def samplemixinmethod(method):
    """Marks a method as being a mixin.

    Adds the '__issamplemixin__' attribute with value True to the decorated function.

    Examples:
        >>> @samplemixinmethod
        >>> def f():
        ...     pass
        >>> f.__issamplemixin__
        True

    """
    # NB: decorator name was chosen to be consistent with @classmethod and @staticmethod
    method.__issamplemixin__ = True
    return method



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
# ================================================================================================
"""
The :class:`.Sampler` abstract base class (`ABC <https://docs.python.org/3.6/library/abc.html#module-abc>`_\ )
helps you create new dimod samplers.

Any new dimod sampler must define a subclass of :class:`.Sampler` that implements
abstract properties :attr:`~.Sampler.parameters` and :attr:`~.Sampler.properties`
and one of the abstract methods :meth:`~.Sampler.sample`, :meth:`~.Sampler.sample_ising`,
or :meth:`~.Sampler.sample_qubo`. The :class:`.Sampler` class provides the complementary
methods as mixins and ensures consistent responses.

For example, the following steps show how to easily create a dimod sampler. It is
sufficient to implement a single method (in this example the :meth:`sample_ising` method)
to create a dimod sampler with the :class:`.Sampler` class.

.. code-block:: python

    class LinearIsingSampler(dimod.Sampler):

        def sample_ising(self, h, J):
            sample = linear_ising(h, J)
            energy = dimod.ising_energy(sample, h, J)
            return dimod.Response.from_dicts([sample], {'energy': [energy]})

        @property
        def properties(self):
            return dict()

        @property
        def parameters(self):
            return dict()

For this example, the implemented sampler :meth:`~.Sampler.sample_ising` can be based on
a simple placeholder function, which returns a sample that minimizes the linear terms:

.. code-block:: python

    def linear_ising(h, J):
        sample = {}
        for v in h:
            if h[v] < 0:
                sample[v] = +1
            else:
                sample[v] = -1
        return sample


The :class:`.Sampler` ABC provides the other sample methods "for free"
as mixins.

.. code-block:: python

    sampler = LinearIsingSampler()
    response = sampler.sample_ising({'a': -1}, {})  # Implemented by class LinearIsingSampler
    response = sampler.sample_qubo({('a', 'a'): 1})  # Mixin provided by Sampler class
    response = sampler.sample(BinaryQuadraticModel.from_ising({'a': -1}, {}))  # Mixin provided by Sampler class

Below is a more complex version of the same sampler, where the :attr:`properties` and
:attr:`parameters` properties return non-empty dicts.

.. code-block:: python

    class FancyLinearIsingSampler(dimod.Sampler):
        def __init__(self):
            self._properties = {'description': 'a simple sampler that only considers the linear terms'}
            self._parameters = {'verbose': []}

        def sample_ising(self, h, J, verbose=False):
            sample = linear_ising(h, J)
            energy = dimod.ising_energy(sample, h, J)
            if verbose:
                print(sample)
            return dimod.Response.from_dicts([sample], {'energy': [energy]})

        @property
        def properties(self):
            return self._properties

        @property
        def parameters(self):
            return self._parameters


"""
import abc

from six import add_metaclass

from dimod.binary_quadratic_model import BinaryQuadraticModel
from dimod.exceptions import InvalidSampler
from dimod.meta import SamplerABCMeta, samplemixinmethod
from dimod.vartypes import Vartype


__all__ = ['Sampler']


@add_metaclass(SamplerABCMeta)
class Sampler:
    """Abstract base class for dimod samplers.

    Provides all methods :meth:`~.Sampler.sample`, :meth:`~.Sampler.sample_ising`,
    :meth:`~.Sampler.sample_qubo` assuming at least one is implemented.

    """

    @abc.abstractproperty  # for python2 compatibility
    def parameters(self):
        """dict: A dict where keys are the keyword parameters accepted by the sampler
        methods and values are lists of the properties relevent to each parameter.

        Examples:
            This example adds a `verbose` parameter to an Ising sampler.

            .. code-block:: python

                class IsingSampler(dimod.Sampler):
                    def __init__(self):
                        self._parameters = {'verbose': []}

                    def sample_ising(self, h, J, verbose=False):
                        sample = linear_ising(h, J) # Implemented elsewhere
                        energy = dimod.ising_energy(sample, h, J)
                        if verbose:
                            print(sample)
                        return dimod.Response.from_dicts([sample], {'energy': [energy]})

                    @property
                    def properties(self):
                        return dict()

                    @property
                    def parameters(self):
                        return self._parameters

        """
        pass

    @abc.abstractproperty  # for python2 compatibility
    def properties(self):
        """dict: A dict containing any additional information about the sampler.

        Examples:
            This example adds a `description` property to an Ising sampler.

            .. code-block:: python

                class IsingSampler(dimod.Sampler):
                    def __init__(self):
                        self._properties = {'description': 'an example Ising sampler'}

                    def sample_ising(self, h, J):
                        sample = linear_ising(h, J) # Implemented elsewhere
                        energy = dimod.ising_energy(sample, h, J)
                        return dimod.Response.from_dicts([sample], {'energy': [energy]})

                    @property
                    def properties(self):
                        return self._properties

                    @property
                    def parameters(self):
                        return dict()

        """
        pass

    @samplemixinmethod
    def sample(self, bqm, **parameters):
        """Samples from a binary quadratic model using an implemented sample method.
        """
        if bqm.vartype is Vartype.SPIN:
            Q, offset = bqm.to_qubo()
            response = self.sample_qubo(Q, **parameters)
            response.change_vartype(Vartype.SPIN, data_vector_offsets={'energy': offset})
            return response
        elif bqm.vartype is Vartype.BINARY:
            h, J, offset = bqm.to_ising()
            response = self.sample_ising(h, J, **parameters)
            response.change_vartype(Vartype.BINARY, data_vector_offsets={'energy': offset})
            return response
        else:
            raise RuntimeError("binary quadratic model has an unknown vartype")

    @samplemixinmethod
    def sample_ising(self, h, J, **parameters):
        """Samples from an Ising model using an implemented sample method.
        """
        bqm = BinaryQuadraticModel.from_ising(h, J)
        response = self.sample(bqm, **parameters)
        return response

    @samplemixinmethod
    def sample_qubo(self, Q, **parameters):
        """Samples from a QUBO using an implemented sample method.
        """
        bqm = BinaryQuadraticModel.from_qubo(Q)
        response = self.sample(bqm, **parameters)
        return response



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
# ================================================================================================
"""
Enumeration of valid variable types for binary quadratic models.

Examples:
    This example shows easy access to different Vartypes, which are in the main
    namespace.

    >>> vartype = dimod.SPIN
    >>> print(vartype)
    Vartype.SPIN
    >>> vartype = dimod.BINARY
    >>> print(vartype)
    Vartype.BINARY
    >>> vartype = dimod.Vartype.SPIN
    >>> print(vartype)
    Vartype.SPIN
    >>> isinstance(vartype, dimod.Vartype)
    True

    This example shows access by value or name.

    >>> print(dimod.Vartype({0, 1}))
    Vartype.BINARY
    >>> print(dimod.Vartype['SPIN'])
    Vartype.SPIN

    This example uses the `.value` parameter to validate.

    >>> sample = {'u': -1, 'v': 1}
    >>> vartype = dimod.Vartype.SPIN
    >>> all(val in vartype.value for val in sample.values())
    True

"""
import enum

__all__ = ['Vartype', 'SPIN', 'BINARY']


class Vartype(enum.Enum):
    """An :py:class:`~enum.Enum` over the types of variables for the binary quadratic model.

    Attributes:
        SPIN (:class:`.Vartype`): Vartype for spin-valued models; variables of
           the model are either -1 or 1.
        BINARY (:class:`.Vartype`): Vartype for binary models; variables of the
           model are either 0 or 1.

    """
    SPIN = frozenset({-1, 1})
    BINARY = frozenset({0, 1})


SPIN = Vartype.SPIN
BINARY = Vartype.BINARY



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
# ================================================================================================

from __future__ import absolute_import

from dimod.core import *
import dimod.core

from dimod.reference import *
import dimod.reference

import dimod.testing

from dimod.binary_quadratic_model import *
import dimod.binary_quadratic_model

import dimod.decorators

# for now, let's just import the most important embedding functions
from dimod.embedding import embed_bqm, embed_ising, embed_qubo, iter_unembed, chain_break_frequency, unembed_response
import dimod.embedding

from dimod.exceptions import *
import dimod.exceptions

from dimod.higherorder import *
import dimod.higherorder

from dimod.package_info import __version__, __author__, __authoremail__, __description__

from dimod.response import *
import dimod.response

from dimod.utilities import *
import dimod.utilities

from dimod.vartypes import *
import dimod.vartypes



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
# ================================================================================================

from __future__ import absolute_import

import sys
import os

from setuptools import setup

# add __version__, __author__, __authoremail__, __description__ to this namespace
_PY2 = sys.version_info.major == 2
my_loc = os.path.dirname(os.path.abspath(__file__))
os.chdir(my_loc)
if _PY2:
    execfile(os.path.join(".", "dimod", "package_info.py"))
else:
    exec(open(os.path.join(".", "dimod", "package_info.py")).read())

install_requires = ['numpy>=1.11.3,<2.0.0',
                    'six>=1.10.0,<2.0.0',
                    'jsonschema>=2.6.0,<3.0.0']

extras_require = {'all': ['networkx>=2.0,<3.0',
                          'pandas>=0.22.0,<0.23.0'],
                  ':python_version == "2.7"': ['futures'],
                  ':python_version <= "3.3"': ['enum34>=1.1.6,<2.0.0']}

packages = ['dimod',
            'dimod.core',
            'dimod.embedding',
            'dimod.io',
            'dimod.reference',
            'dimod.reference.composites',
            'dimod.reference.samplers',
            'dimod.testing']

classifiers = [
    'License :: OSI Approved :: Apache Software License',
    'Operating System :: OS Independent',
    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    ]

python_requires = '>=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*,!=3.4.*'

setup(
    name='dimod',
    version=__version__,
    author=__author__,
    author_email=__authoremail__,
    description=__description__,
    long_description=open('README.rst').read(),
    url='https://github.com/dwavesystems/dimod',
    download_url='https://github.com/dwavesystems/dimod/releases',
    license='Apache 2.0',
    packages=packages,
    install_requires=install_requires,
    extras_require=extras_require,
    include_package_data=True,
    classifiers=classifiers,
    python_requires=python_requires
)



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
# ================================================================================================
"""

A structured sampler can only sample from binary quadratic models with a specific graph.

For structured samplers you must implement the :attr:`~.Structured.nodelist`
and :attr:`~.Structured.edgelist` properties. The :class:`.Structured` abstract base class
provides access to the :attr:`~.Structured.structure` and :attr:`~.Structured.adjacency`
properties as well as any method or properties required by the :class:`.Sampler` abstract
base class. The :obj:`~.bqm_structured` decorator verifies that any given binary quadratic
model conforms to the supported structure.

Examples:
    This simple example shows a structured sampler that can only sample from a binary quadratic model
    with two variables and one interaction.

    .. code-block:: python

        class TwoVariablesSampler(dimod.Sampler, dimod.Structured):
            @property
            def nodelist(self):
                return [0, 1]

            @property
            def edgelist(self):
                return [(0, 1)]

            @property
            def properties(self):
                return dict()

            @property
            def parameters(self):
                return dict()

            @dimod.decorators.bqm_structured
            def sample(self, bqm):
                # All bqm's passed in will be a subgraph of the sampler's structure
                variable_list = list(bqm.linear)
                samples = []
                energies = []
                for values in itertools.product(bqm.vartype.value, repeat=len(bqm)):
                    sample = dict(zip(variable_list, values))
                    samples.append(sample)
                    energies.append(bqm.energy(sample))

                response = dimod.Response.from_dicts(samples, {'energy': energies}, vartype=bqm.vartype)

                return response

"""
import abc

from collections import namedtuple

from six import add_metaclass

__all__ = ['Structured']

_Structure = namedtuple("Structure", ['nodelist', 'edgelist', 'adjacency'])


@add_metaclass(abc.ABCMeta)
class Structured:
    """The abstract base class for dimod structured samplers.

    Provides the :attr:`.Structured.adjacency` and :attr:`.Structured.structure` properties.

    Abstract properties :attr:`~.Structured.nodelist` and :attr:`~.Structured.edgelist`
    must be implemented.

    """
    @abc.abstractproperty
    def nodelist(self):
        """list: Nodes/variables allowed by the sampler formatted as a list.

        Examples:
            This example defines the `nodelist` for a structured sampler that can only sample
            from a binary quadratic model with two variables.

            .. code-block:: python

                class TwoVariablesSampler(dimod.Sampler, dimod.Structured):
                    @property
                    def nodelist(self):
                        return [0, 1]

                    # Remaining properties and code for the sampler

        """
        pass

    @abc.abstractproperty
    def edgelist(self):
        """list[(node, node)]: Edges/interactions allowed by the sampler, formatted as
        a list where each edge/interaction is a 2-tuple.

        Examples:
            This example defines the `edgelist` for a structured sampler that can only sample
            from a binary quadratic model with a single interaction between two variables.

            .. code-block:: python

                class TwoVariablesSampler(dimod.Sampler, dimod.Structured):
                    @property
                    def edgelist(self):
                        return [(0, 1)]

                    # Remaining properties and code for the sampler

        """
        pass

    @property
    def adjacency(self):
        """dict[variable, set]: Adjacency structure formatted as a dict, where keys are the
        nodes of the structured sampler and values are sets of all adjacent nodes for
        each key node.

        Examples:
            This example shows the adjacencies for a placeholder structured sampler that
            samples only from the K4 complete graph, where each of the four nodes connects
            to all the other nodes.

            >>> class K4StructuredClass(dimod.Structured):
            ...     @property
            ...     def nodelist(self):
            ...         return [1, 2, 3, 4]
            ...
            ...     @property
            ...     def edgelist(self):
            ...         return [(1, 2), (1, 3), (1, 4), (2, 3), (2, 4), (3, 4)]
            >>> K4sampler = K4StructuredClass()
            >>> K4sampler.adjacency.keys()
            [1, 2, 3, 4]

        """
        if not hasattr(self, '_adjacency'):
            adjacency = {v: set() for v in self.nodelist}
            for u, v in self.edgelist:
                if v in adjacency[u]:
                    raise ValueError("Each edge in edgelist must be unique")
                adjacency[u].add(v)
                adjacency[v].add(u)
            self._adjacency = adjacency
            return adjacency
        return self._adjacency

    @property
    def structure(self):
        """Structure of the structured sampler formatted as a :class:`~collections.namedtuple`
        :samp:`Structure(nodelist, edgelist, adjacency)`, where the 3-tuple values are
        the :attr:`~.Structured.nodelist` and :attr:`~.Structured.edgelist` properties
        and :meth:`~.Structured.adjacency` method.

        Examples:
            This example shows the structure of a placeholder structured sampler that
            samples only from the K3 complete graph, where each of the three nodes connects
            to all the other nodes.

            >>> class K3StructuredClass(dimod.Structured):
            ...     @property
            ...     def nodelist(self):
            ...         return [1, 2, 3]
            ...
            ...     @property
            ...     def edgelist(self):
            ...         return [(1, 2), (1, 3), (2, 3)]
            >>> K3sampler = K3StructuredClass()
            >>> K3sampler.structure.edgelist
            [(1, 2), (1, 3), (2, 3)]

        """
        return _Structure(self.nodelist, self.edgelist, self.adjacency)
